"""Created on Friday October 01 10:09:11 2021

In this module you will find useful functions and classes to operate with data
recorded using spellers based on code-modulated visual evoked potentials
(c-VEP), which are widely used by the BCI community. Enjoy!

@author: Víctor Martínez-Cagigal
"""
import medusa as mds
from medusa import components
from medusa import meeg
from medusa import spatial_filtering as sf
from medusa import epoching as ep
import copy, warnings
import itertools

from abc import ABC, abstractmethod
import numpy as np
from tqdm import tqdm

LFSR_PRIMITIVE_POLYNOMIALS = \
    {
        'base': {
            2: {
                'order': {
                    2: [1, 1],
                    3: [1, 0, 1],
                    4: [1, 0, 0, 1],
                    5: [0, 1, 0, 0, 1],
                    6: [0, 0, 0, 0, 1, 1],
                    7: [0, 0, 0, 0, 0, 1, 1],
                    8: [1, 1, 0, 0, 0, 0, 1, 1],
                    9: [0, 0, 0, 1, 0, 0, 0, 0, 1],
                    10: [0, 0, 1, 0, 0, 0, 0, 0, 0, 1],
                    11: [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1],
                    12: [0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1],
                    13: [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1],
                    14: [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1],
                    15: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1],
                    16: [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1],
                    17: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    18: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                    19: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1,
                         1],
                    20: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                         0, 0],
                    21: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 1, 0],
                    22: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 1],
                    23: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         1, 0, 0, 0, 0],
                    24: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                         0, 0, 0, 0, 1, 1],
                    25: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 1, 0, 0],
                    26: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 1, 0, 0, 0, 1, 1],
                    27: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 1, 0, 0, 1, 1],
                    28: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    29: [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    30: [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1]
                }
            },
            3: {
                'order': {
                    2: [2, 1],
                    3: [0, 1, 2],
                    4: [0, 0, 2, 1],
                    5: [0, 0, 0, 1, 2],
                    6: [0, 0, 0, 0, 2, 1],
                    7: [0, 0, 0, 0, 2, 1, 2],
                }
            },
            5: {
                'order': {
                    2: [4, 3],
                    3: [0, 2, 3],
                    4: [0, 4, 3, 3],
                }
            },
            7: {
                'order': {
                    2: [1, 4]
                }
            },
            11: {
                'order': {
                    2: [1, 3]
                }
            },
            13: {
                'order': {
                    2: [1, 11]
                }
            }
        }
    }


# --------------------------- c-VEP DATA MANAGEMENT -------------------------- #
class CVEPSpellerData(components.ExperimentData):
    """Experiment info class for c-VEP-based spellers. It supports nested
    multi-level paradigms. This unified class can be used to represent a run
    of every c-VEP stimulation paradigm designed to date, and is the expected
    class for feature extraction and command decoding functions of the module
    medusa.bci.cvep_paradigms. It is complicated, but powerful so.. use it well!
    """

    def __init__(self, mode, paradigm_conf, commands_info, onsets, command_idx,
                 unit_idx, level_idx, matrix_idx, cycle_idx, trial_idx,
                 cvep_model, spell_result, fps_resolution, spell_target=None,
                 raster_events=None, **kwargs):

        # Check errors
        if mode not in ('train', 'test'):
            raise ValueError('Unknown mode. Possible values {train, test}')

        # Standard attributes
        self.mode = mode
        self.paradigm_conf = paradigm_conf
        self.commands_info = commands_info
        self.onsets = onsets
        self.command_idx = command_idx
        self.unit_idx = unit_idx
        self.level_idx = level_idx
        self.matrix_idx = matrix_idx
        self.cycle_idx = cycle_idx
        self.trial_idx = trial_idx
        self.cvep_model = cvep_model
        self.spell_result = spell_result
        self.fps_resolution = fps_resolution
        self.spell_target = spell_target
        self.raster_events = raster_events

        # Optional attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
        return rec_dict

    @classmethod
    def from_serializable_obj(cls, dict_data):
        return cls(**dict_data)


class CVEPSpellerDataset(components.Dataset):
    """ This class inherits from medusa.data_structures.Dataset, increasing
    its functionality for datasets with data from c-VEP-based spellers. It
    provides common ground for the rest of functions in the module.
    """

    def __init__(self, channel_set, fs=None, biosignal_att_key='eeg',
                 experiment_att_key='cvepspellerdata', experiment_mode=None,
                 track_attributes=None):
        """ Constructor

        Parameters
        ----------
        channel_set : meeg.EEGChannelSet
            EEG channel set. Only these channels will be kept in the dataset,
            the others will be discarded. Also, the signals will be rearranged,
            keeping the same channel order, avoiding errors in future stages of
            the signal processing pipeline
        fs : int, float or None
            Sample rate of the recordings. If there are recordings with
            different sample rates, the consistency of the dataset can be
            still assured using resampling
        biosignal_att_key : str
            Name of the attribute containing the target biosignal that will be
            used to extract the features. It has to be the same in all
            recordings (e.g., 'eeg', 'meg').
        experiment_att_key : str or None
            Name of the attribute containing the target experiment that will be
            used to extract the features. It has to be the same in all
            recordings (e.g., 'rcp_data', 'cake_paradigm_data'). It is
            mandatory when a recording of the dataset contains more than 1
            experiment data
        experiment_mode : str {'train'|'test'|None}
            Mode of the experiment. If this dataset will be used to fit a model,
            set to train to avoid errors
        track_attributes: dict of dicts or None
            This parameter indicates custom attributes that must be tracked in
            feature extraction functions and how. The keys are the name of the
            attributes, whereas the values are dicts indicating the tracking
            mode {'concatenate'|'append'} and parent. Option concatenate is
            only available for attributes of type list or numpy arrays,
            forming a 1 dimensional array with the data from all recordings.
            Option append is used to save all kind of objects for each
            recording, forming a list whose length will be the number of
            recordings in the dataset. A set of default attributes is defined,
            so this parameter will be None in most cases. Example to track 2
            custom attributes (i.e., date and experiment_equipment):
                track_attributes = {
                    'date': {
                        'track_mode': 'append',
                        'parent': None
                    },
                    'experiment_equipment': {
                        'track_mode': 'append',
                        'parent': experiment_att_key
                    }
                }
        """
        # Check errors
        if experiment_mode is not None:
            if experiment_mode not in ('train', 'test'):
                raise ValueError('Parameter experiment_mode must be '
                                 '{train|test|None}')

        # Default track attributes
        default_track_attributes = {
            'subject_id': {
                'track_mode': 'append',
                'parent': None
            },
            'paradigm_conf': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'commands_info': {
                'track_mode': 'append',
                'parent': experiment_att_key
            },
            'onsets': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'command_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'unit_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'level_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'matrix_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'cycle_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'trial_idx': {
                'track_mode': 'concatenate',
                'parent': experiment_att_key
            },
            'spell_result': {
                'track_mode': 'append',
                'parent': experiment_att_key
            }
        }

        if experiment_mode == 'train':
            default_track_attributes_train = {
                'cvep_labels': {
                    'track_mode': 'concatenate',
                    'parent': experiment_att_key
                },
                'spell_target': {
                    'track_mode': 'append',
                    'parent': experiment_att_key
                }
            }
            default_track_attributes = {
                **default_track_attributes,
                **default_track_attributes_train
            }

        track_attributes = \
            default_track_attributes if track_attributes is None else \
                {**default_track_attributes, **track_attributes}

        # Class attributes
        self.channel_set = channel_set
        self.fs = fs
        self.biosignal_att_key = biosignal_att_key
        self.experiment_att_key = experiment_att_key
        self.experiment_mode = experiment_mode
        self.track_attributes = track_attributes

        # Consistency checker
        checker = self.__get_consistency_checker()
        super().__init__(consistency_checker=checker)

    def __get_consistency_checker(self):
        """Creates a standard consistency checker for c-VEP datasets

        Returns
        -------
        checker : data_structures.ConsistencyChecker
            Standard consistency checker for ERP feature extraction
        """
        # Create consistency checker
        checker = components.ConsistencyChecker()
        # Check that the biosignal exists
        checker.add_consistency_rule(
            rule='check-attribute',
            rule_params={'attribute': self.biosignal_att_key}
        )
        checker.add_consistency_rule(
            rule='check-attribute-type',
            rule_params={'attribute': self.biosignal_att_key,
                         'type': meeg.EEG}
        )
        # Check channels
        checker.add_consistency_rule(
            rule='check-values-in-attribute',
            rule_params={'attribute': 'channels',
                         'values': self.channel_set.channels},
            parent=self.biosignal_att_key + '.channel_set'
        )
        # Check sample rate
        if self.fs is not None:
            checker.add_consistency_rule(rule='check-attribute-value',
                                         rule_params={'attribute': 'fs',
                                                      'value': self.fs},
                                         parent=self.biosignal_att_key)
        else:
            warnings.warn('Parameter fs is None. The consistency of the '
                          'dataset cannot be assured. Still, you can use '
                          'target_fs parameter for feature extraction '
                          'and everything should be fine.')

        # Check experiment
        checker.add_consistency_rule(
            rule='check-attribute',
            rule_params={'attribute': self.experiment_att_key}
        )
        checker.add_consistency_rule(
            rule='check-attribute-type',
            rule_params={'attribute': self.experiment_att_key,
                         'type': CVEPSpellerData}
        )
        # Check mode
        if self.experiment_mode is not None:
            checker.add_consistency_rule(
                rule='check-attribute-value',
                rule_params={'attribute': 'mode',
                             'value': self.experiment_mode},
                parent=self.experiment_att_key
            )

        # Check track_attributes
        if self.track_attributes is not None:
            for key, value in self.track_attributes.items():
                checker.add_consistency_rule(
                    rule='check-attribute',
                    rule_params={'attribute': key},
                    parent=value['parent']
                )
                if value['track_mode'] == 'concatenate':
                    checker.add_consistency_rule(
                        rule='check-attribute-type',
                        rule_params={'attribute': key,
                                     'type': [list, np.ndarray]},
                        parent=value['parent']
                    )

        return checker

    def custom_operations_on_recordings(self, recording):
        # Select channels
        eeg = getattr(recording, self.biosignal_att_key)
        eeg.change_channel_set(self.channel_set)
        return recording


# ---------------------------------- MODELS ---------------------------------- #
class CVEPModelCircularShifting(components.Algorithm):

    def __init__(self, bpf=[[7, (1.0, 30.0)]], notch=[7, (49.0, 51.0)],
                 art_rej=None, correct_raster_latencies=False,
                 *args, **kwargs):
        super().__init__()

        if len(bpf) == 1:
            if notch is not None:
                self.add_method('prep_method', StandardPreprocessing(
                    bpf_order=bpf[0][0], bpf_cutoff=bpf[0][1],
                    notch_order=notch[0], notch_cutoff=notch[1]))
            else:
                self.add_method('prep_method', StandardPreprocessing(
                    bpf_order=bpf[0][0], bpf_cutoff=bpf[0][1],
                    notch_order=None, notch_cutoff=None))
        else:
            filter_bank = []
            for i in range(len(bpf)):
                filter_bank.append({
                    'order': bpf[i][0],
                    'cutoff': bpf[i][1],
                    'btype': 'bandpass'
                })
            if notch is not None:
                self.add_method('prep_method', FilterBankPreprocessing(
                    filter_bank=filter_bank, notch_order=notch[0],
                    notch_cutoff=notch[1]))
            else:
                self.add_method('prep_method', FilterBankPreprocessing(
                    filter_bank=filter_bank, notch_order=None,
                    notch_cutoff=None))

        # Feature extraction and classification (circular shifting)
        self.add_method('clf_method', CircularShiftingClassifier(
            art_rej=art_rej,
            correct_raster_latencies=correct_raster_latencies
        ))

        # Early stopping
        self.add_method('es_method', CircularShiftingEarlyStopping())

    def check_predict_feasibility(self, dataset):
        return self.get_inst('clf_method')._is_predict_feasible(dataset)

    def check_predict_feasibility_signal(self, times, onsets, fs):
        return self.get_inst('clf_method')._is_predict_feasible_signal(
            times, onsets, fs)

    def fit_dataset(self, dataset, **kwargs):
        # Safe copy
        data = copy.deepcopy(dataset)

        # Preprocessing
        data = self.get_inst('prep_method').fit_transform_dataset(
            dataset=data,
            show_progress_bar=True
        )

        # Feature extraction and classification
        fitted_info = self.get_inst('clf_method').fit_dataset(
            dataset=data,
            std_epoch_rejection=3.0,
            show_progress_bar=True
        )

        return fitted_info

    def predict_dataset(self, dataset):
        # Safe copy
        data = copy.deepcopy(dataset)

        # Preprocessing
        data = self.get_inst('prep_method').transform_dataset(
            dataset=data,
            show_progress_bar=True
        )

        # Feature extraction and classification
        pred_items = self.get_inst('clf_method').predict_dataset(
            dataset=data,
            show_progress_bar=True
        )

        # Extract the selected items using the maximum number of cycles
        selected_seq = 0  # todo: several sequences in the same matrix
        spell_result = []
        for item in pred_items:
            spell_result.append(
                item[-1][selected_seq]['sorted_cmds'][0]['label'])

        # Extract the selected items depending on the number of cycles
        spell_result_per_cycle = []
        for item in pred_items:
            trial_result_per_cycle = {}
            for nc, pred in enumerate(item):
                trial_result_per_cycle[nc] = pred[selected_seq][
                    'sorted_cmds'][0]['label']
            spell_result_per_cycle.append(trial_result_per_cycle)

        # Create the decoding dictionary
        cmd_decoding = {
            'spell_result': spell_result,
            'spell_result_per_cycle': spell_result_per_cycle,
            'items_by_no_cycle': pred_items
        }
        return cmd_decoding

    def predict(self, times, signal, trial_idx, exp_data, sig_data):
        # Safe copy
        times_ = copy.deepcopy(times)
        signal_ = copy.deepcopy(signal)
        trial_idx_ = copy.deepcopy(trial_idx)
        exp_data_ = copy.deepcopy(exp_data)
        sig_data_ = copy.deepcopy(sig_data)

        # Preprocessing
        signal_ = self.get_inst('prep_method').transform_signal(
            signal=signal_
        )

        # Feature extraction and classification
        pred_item_by_no_cycles = self.get_inst('clf_method').predict(
            times_, signal_, trial_idx_, exp_data_, sig_data_
        )

        # Extract the selected label using the maximum number of cycles
        selected_seq = 0  # todo: several sequences in the same matrix
        spell_result = pred_item_by_no_cycles[-1][selected_seq][
            'sorted_cmds'][0]['label']

        # Extract the selected label depending on the number of cycles
        spell_result_per_cycle = {}
        for nc, pred in enumerate(pred_item_by_no_cycles):
            spell_result_per_cycle[nc] = pred[selected_seq]['sorted_cmds'][
                0]['label']

        # Create the decoding dictionary
        cmd_decoding = {
            'spell_result': spell_result,
            'spell_result_per_cycle': spell_result_per_cycle,
            'items_by_no_cycle': pred_item_by_no_cycles
        }

        return cmd_decoding

    def must_stop(self, corr_vector, std=3.0):
        # Safe copy
        corr_vector_ = copy.deepcopy(corr_vector)
        return self.get_inst('es_method').check_early_stop(
            corr_vector=corr_vector,
            std=std
        )


# ------------------------------- ALGORITHMS -------------------------------- #
class StandardPreprocessing(components.ProcessingMethod):
    """Just the common preprocessing applied in c-VEP-based spellers. Simple,
    quick and effective: frequency IIR band-pass and notch filters
    """

    def __init__(self, bpf_order=7, bpf_cutoff=(0.5, 60.0), notch_order=7,
                 notch_cutoff=(49.0, 51.0)):
        super().__init__(fit_transform_signal=['signal'],
                         fit_transform_dataset=['dataset'])
        # Parameters
        self.bpf_order = bpf_order
        self.bpf_cutoff = bpf_cutoff
        self.notch_order = notch_order
        self.notch_cutoff = notch_cutoff
        self.filt_method = 'sosfiltfilt'

        # Variables
        self.bpf_iir_filter = None
        self.notch_iir_filter = None

    def fit(self, fs):
        """Fits the IIR filters.

        Parameters
        ----------
        fs: float
            Sample rate of the signal.
        """
        # Bandpass
        self.bpf_iir_filter = mds.IIRFilter(order=self.bpf_order,
                                            cutoff=self.bpf_cutoff,
                                            btype='bandpass',
                                            filt_method=self.filt_method)
        self.bpf_iir_filter.fit(fs)
        # Notch
        if self.notch_cutoff is not None:
            self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
                                                  cutoff=self.notch_cutoff,
                                                  btype='bandstop',
                                                  filt_method=self.filt_method)
            self.notch_iir_filter.fit(fs)

    def transform_signal(self, signal):
        """Transforms an EEG signal applying IIR filterings

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        """
        signal = self.bpf_iir_filter.transform(signal)
        if self.notch_iir_filter is not None:
            signal = self.notch_iir_filter.transform(signal)
        return signal

    def fit_transform_signal(self, signal, fs):
        """Fits the IIR filters and transforms an EEG signal applying them
        sequentially

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        fs: float
            Sample rate of the signal.
        """
        self.fit(fs)
        signal = self.transform_signal(signal)
        return signal

    def fit_transform_dataset(self, dataset: CVEPSpellerDataset,
                              show_progress_bar=True):
        """Fits the IIR filters and transforms an EEG dataset applying the
        filters sequentially. Each recording is preprocessed independently,
        taking into account possible differences in sample rate.

        Parameters
        ----------
        dataset: ERPSpellerDataset
            ERPSpellerDataset with the recordings to be preprocessed.
        show_progress_bar: bool
            Show progress bar
        """
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Fitting preprocessing')
        for rec in dataset.recordings:
            eeg = getattr(rec, dataset.biosignal_att_key)
            eeg.signal = self.fit_transform_signal(eeg.signal, eeg.fs)
            setattr(rec, dataset.biosignal_att_key, eeg)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return dataset

    def transform_dataset(self, dataset: CVEPSpellerDataset,
                          show_progress_bar=True):
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Applying preprocessing')
        for rec in dataset.recordings:
            eeg = getattr(rec, dataset.biosignal_att_key)
            eeg.signal = self.transform_signal(eeg.signal)
            setattr(rec, dataset.biosignal_att_key, eeg)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return dataset


class FilterBankPreprocessing(components.ProcessingMethod):
    """Just the common preprocessing applied in c-VEP-based spellers. Simple,
    quick and effective: frequency IIR band-pass and notch filters
    """

    def __init__(self, filter_bank=None, notch_order=7,
                 notch_cutoff=(49.0, 51.0)):
        super().__init__(fit_transform_signal=['signal'],
                         fit_transform_dataset=['dataset'])
        if filter_bank is None:
            filter_bank = [{'order': 7,
                            'cutoff': (8.0, 60.0),
                            'btype': 'bandpass'},
                           {'order': 7,
                            'cutoff': (12.0, 60.0),
                            'btype': 'bandpass'},
                           {'order': 7,
                            'cutoff': (30.0, 60.0),
                            'btype': 'bandpass'},
                           ]

        # Error check
        if not filter_bank:
            raise ValueError('[FilterBankPreprocessing] Filter bank parameter '
                             '"filter_bank" must be a list containing all '
                             'necessary information to perform the filtering!')
        for filter in filter_bank:
            if not isinstance(filter, dict):
                raise ValueError('[FilterBankPreprocessing] Each filter must '
                                 'be a dict()!')
            if 'order' not in filter or \
                    'cutoff' not in filter or \
                    'btype' not in filter:
                raise ValueError('[FilterBankPreprocessing] Each filter must '
                                 'be a dict() containing the following keys: '
                                 '"order", "cutoff" and "btype"!')

        # Parameters
        self.filter_bank = filter_bank
        self.notch_order = notch_order
        self.notch_cutoff = notch_cutoff
        self.filt_method = 'sosfiltfilt'

        # Variables
        self.filter_bank_iir_filters = None
        self.notch_iir_filter = None

    def fit(self, fs):
        """Fits the IIR filters.

        Parameters
        ----------
        fs: float
            Sample rate of the signal.
        """
        # Filter bank
        self.filter_bank_iir_filters = []
        for filter in self.filter_bank:
            iir = mds.IIRFilter(order=filter['order'],
                                cutoff=filter['cutoff'],
                                btype=filter['btype'],
                                filt_method=self.filt_method)
            iir.fit(fs)
            self.filter_bank_iir_filters.append(iir)

        # Notch
        if self.notch_cutoff is not None:
            self.notch_iir_filter = mds.IIRFilter(order=self.notch_order,
                                                  cutoff=self.notch_cutoff,
                                                  btype='bandstop',
                                                  filt_method=self.filt_method)
            self.notch_iir_filter.fit(fs)

    def transform_signal(self, signal):
        """Transforms an EEG signal applying the filter bank

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        """
        if self.notch_iir_filter is not None:
            signal = self.notch_iir_filter.transform(signal)
        signals = []
        for filter in self.filter_bank_iir_filters:
            signal_ = signal.copy()
            signals.append(filter.transform(signal_))
        return signals

    def fit_transform_signal(self, signal, fs):
        """Fits the IIR filters and transforms an EEG signal applying them
        sequentially

        Parameters
        ----------
        signal: np.array or list
            Signal to transform. Shape [n_samples x n_channels]
        fs: float
            Sample rate of the signal.
        """
        self.fit(fs)
        signals = self.transform_signal(signal)
        return signals

    def fit_transform_dataset(self, dataset: CVEPSpellerDataset,
                              show_progress_bar=True):
        """Fits the IIR filters and transforms an EEG dataset applying the
        filters sequentially. Each recording is preprocessed independently,
        taking into account possible differences in sample rate.

        Parameters
        ----------
        dataset: ERPSpellerDataset
            ERPSpellerDataset with the recordings to be preprocessed.
        show_progress_bar: bool
            Show progress bar
        """
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Fitting preprocessing')
        for rec in dataset.recordings:
            eeg = getattr(rec, dataset.biosignal_att_key)
            eeg.signal = self.fit_transform_signal(eeg.signal, eeg.fs)
            setattr(rec, dataset.biosignal_att_key, eeg)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return dataset

    def transform_dataset(self, dataset: CVEPSpellerDataset,
                          show_progress_bar=True):
        pbar = None
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Applying preprocessing')
        for rec in dataset.recordings:
            eeg = getattr(rec, dataset.biosignal_att_key)
            eeg.signal = self.transform_signal(eeg.signal)
            setattr(rec, dataset.biosignal_att_key, eeg)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return dataset


class CircularShiftingClassifier(components.ProcessingMethod):
    """Standard feature extraction method for c-VEP-based spellers. Basically,
    it computes a template for each sequence.
    """

    def __init__(self, correct_raster_latencies=False, art_rej=None, **kwargs):
        """ Class constructor """
        super().__init__(fit_dataset=['templates',
                                      'cca_by_seq'])
        self.fitted = dict()

        self.art_rej = art_rej
        self.correct_raster_latencies = correct_raster_latencies

    def _assert_consistency(self, dataset: CVEPSpellerDataset):
        len_seqs = set()
        fs = set()
        fps_resolution = set()
        unique_seqs_by_run = []
        unique_all_seqs = set()
        is_filter_bank = []
        for rec in dataset.recordings:
            rec_sig = getattr(rec, dataset.biosignal_att_key)
            rec_exp = getattr(rec, dataset.experiment_att_key)
            if rec_exp.mode != 'train':
                raise ValueError('There is at least one CVEPSpellerData '
                                 'instance that was not recording under train '
                                 'mode. Aborting feature extraction...')
            if not hasattr(rec_exp, 'spell_target'):
                raise ValueError('There is at least one CVEPSpellerData '
                                 'instance that has not "spell_target" data. '
                                 'Aborting feature extraction...')
            fps_resolution.add(rec_exp.fps_resolution)
            fs.add(rec_sig.fs)
            unique_seqs = get_unique_sequences_from_targets(rec_exp)
            for seq_ in unique_seqs:
                len_seqs.add(len(seq_))
                unique_all_seqs.add(seq_)
            unique_seqs_by_run.append(unique_seqs)

            if isinstance(rec_sig.signal, list):
                is_filter_bank.append(True)
            else:
                is_filter_bank.append(False)
        if len(len_seqs) > 1:
            raise ValueError('There are sequences with different lengths in '
                             'the CVEPSpellerDataset instance! Aborting feature'
                             ' extraction...')
        if len(fs) > 1:
            raise ValueError('There are CVEPSpellerData instances with '
                             'different sampling rates! Aborting feature '
                             'extraction...')
        if len(fps_resolution) > 1:
            raise ValueError('There are CVEPSpellerData instances with '
                             'different refresh rates! Aborting feature '
                             'extraction...')
        if len(unique_all_seqs) > 1:
            # Check if some sequences are shifted versions of another
            unique_ = list(unique_all_seqs)
            all_combos = np.array(list(itertools.combinations(
                np.arange(0, len(unique_)), 2)))
            for i_comb in range(all_combos.shape[0]):
                s1 = unique_[all_combos[i_comb][0]]
                s2 = unique_[all_combos[i_comb][1]]
                if check_if_shifted(s1, s2):
                    raise ValueError('There are targets that share shifted '
                                     'versions of the same sequence. Solve that'
                                     'before extracting features!')

        if len(np.unique(is_filter_bank)):
            is_filter_bank = is_filter_bank[0]
        else:
            raise ValueError('There are recordings that have filter banks and '
                             'other do not.')
        len_seq = len_seqs.pop()
        fs = fs.pop()
        fps_resolution = fps_resolution.pop()

        return fs, fps_resolution, len_seq, unique_seqs_by_run, is_filter_bank

    def fit_dataset(self, dataset: CVEPSpellerDataset, std_epoch_rejection=3.0,
                    show_progress_bar=True):

        # Error checking
        fs, fps_resolution, len_seq, unique_seqs_by_run, is_filter_bank = \
            self._assert_consistency(dataset)

        # Init progress bar for sequences
        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Extracting unique sequences')

        # Compute sequence length in milliseconds
        len_epoch_ms = len_seq / fps_resolution * 1000
        len_epoch_sam = int(len_seq / fps_resolution * fs)

        # Get the epochs of each sequence
        epochs_by_seq = {}
        for rec_idx, rec in enumerate(dataset.recordings):
            # Extract recording experiment and biosignal
            rec_exp = getattr(rec, dataset.experiment_att_key)
            rec_sig = getattr(rec, dataset.biosignal_att_key)

            # Filter bank init
            if not is_filter_bank:
                rec_sig.signal = [rec_sig.signal]

            # Get unique sequences for this run
            unique_seqs = unique_seqs_by_run[rec_idx]

            # For each filter bank
            for filter_idx, signal in enumerate(rec_sig.signal):

                # Extract epochs
                epochs = mds.get_epochs_of_events(timestamps=rec_sig.times,
                                                  signal=signal,
                                                  onsets=rec_exp.onsets,
                                                  fs=fs,
                                                  w_epoch_t=[0, len_epoch_ms],
                                                  w_baseline_t=None,
                                                  norm=None)

                # Organize epochs by sequence
                for seq_, ep_idxs_ in unique_seqs.items():
                    if tuple(seq_) not in epochs_by_seq:
                        epochs_by_seq[tuple(seq_)] = \
                            [None for i in range(len(rec_sig.signal))]
                        epochs_by_seq[tuple(seq_)][filter_idx] = \
                            epochs[ep_idxs_, :, :]
                    else:
                        if epochs_by_seq[tuple(seq_)][filter_idx] is None:
                            epochs_by_seq[tuple(seq_)][filter_idx] = \
                                epochs[ep_idxs_, :, :]
                        else:
                            epochs_by_seq[tuple(seq_)][
                                filter_idx] = np.concatenate((
                                epochs_by_seq[tuple(seq_)][filter_idx],
                                epochs[ep_idxs_, :, :]
                            ), axis=0)

            if show_progress_bar:
                pbar.update(1)

        # Precompute nearest channels for online artifact rejection
        sorted_dist_ch = rec_sig.channel_set.sort_nearest_channels()

        # New bar
        if show_progress_bar:
            pbar.close()
            pbar = tqdm(total=len(epochs_by_seq) * len_seq,
                        desc='Creating templates')

        # For each sequence
        seq_dict = dict()
        discarded_epochs = 0
        total_epochs = 0
        for seq_ in epochs_by_seq:

            # For each filter of the bank
            for filter_idx in range(len(epochs_by_seq[seq_])):

                # Offline artifact rejection
                if std_epoch_rejection is not None:
                    epochs_std = np.std(epochs_by_seq[seq_][filter_idx],
                                        axis=1)  # STD per samples
                    ch_std = np.std(epochs_std, axis=0)  # Variation of epochs
                    # For each channel, check if the variation is adequate
                    epoch_to_keep = np.zeros(epochs_std.shape)
                    for i in range(len(ch_std)):
                        epoch_to_keep[:, i] = (
                                (epochs_std[:, i] < (
                                            np.median(epochs_std[:, i]) +
                                            std_epoch_rejection * ch_std[
                                                i])) &
                                (epochs_std[:, i] > (
                                            np.median(epochs_std[:, i]) -
                                            std_epoch_rejection * ch_std[
                                                i]))
                        )
                    # Keep only epochs that are suitable for all channels
                    idx_to_keep = (
                            np.sum(epoch_to_keep, axis=1) == epochs_std.shape[1]
                    )
                    epochs_by_seq[seq_][filter_idx] = \
                        epochs_by_seq[seq_][filter_idx][idx_to_keep, :, :]
                    discarded_epochs += np.sum(idx_to_keep == False)
                    total_epochs += len(idx_to_keep)

                # Canonical Correlation Analysis
                cca = sf.CCA()

                # Reference (main template repeated 'no_cycles' times)
                main_template = np.mean(epochs_by_seq[seq_][filter_idx], axis=0)
                R = np.tile(main_template.T,
                            epochs_by_seq[seq_][filter_idx].shape[0]).T

                # Input data (concatenated epochs)
                X = epochs_by_seq[seq_][filter_idx]
                X = X.reshape((X.shape[0] * X.shape[1], X.shape[2]))

                # Fit CCA and project the main template
                cca.fit(X, R)
                main_template = cca.project(main_template, filter_idx=0,
                                            projection='Wy')
                # Create all possible template shifts
                templates = dict()
                for lag in range(len(seq_)):
                    lag_samples = int(np.round(lag / fps_resolution * fs))
                    lagged_seq = np.roll(seq_, -lag, axis=0)
                    lagged_template = np.roll(main_template, -lag_samples,
                                              axis=0)
                    templates[tuple(lagged_seq)] = lagged_template

                    if show_progress_bar:
                        pbar.update(1)

                # STD by channel (useful for online artifact rejection)
                std_by_channel = np.std(X, axis=0)

                # Store data of each trained sequence
                if tuple(seq_) not in seq_dict:
                    seq_dict[tuple(seq_)] = []
                seq_dict[tuple(seq_)].append(
                    {'cca': cca,
                     'templates': templates,
                     'std_by_channel': std_by_channel,
                     }
                )

        # Store fitted params
        self.fitted = {'sequences': seq_dict,
                       'fs': fs,
                       'fps_resolution': fps_resolution,
                       'len_epoch_ms': len_epoch_ms,
                       'len_epoch_sam': len_epoch_sam,
                       'std_epoch_rejection': std_epoch_rejection,
                       'no_discarded_epochs': discarded_epochs,
                       'no_total_epochs': total_epochs,
                       'sorted_dist_ch': sorted_dist_ch
                       }
        if show_progress_bar:
            pbar.close()

        return self.fitted

    def _is_predict_feasible(self, dataset):
        l_ms = self.fitted['len_epoch_ms']
        for rec in dataset.recordings:
            rec_sig = getattr(rec, dataset.biosignal_att_key)
            rec_exp = getattr(rec, dataset.experiment_att_key)
            feasible = ep.check_epochs_feasibility(timestamps=rec_sig.times,
                                                   onsets=rec_exp.onsets,
                                                   fs=rec_sig.fs,
                                                   t_window=[0, l_ms])
            if feasible != 'ok':
                return False
        return True

    def _is_predict_feasible_signal(self, times, onsets, fs):
        l_ms = self.fitted['len_epoch_ms']
        feasible = ep.check_epochs_feasibility(timestamps=times,
                                               onsets=onsets,
                                               fs=fs,
                                               t_window=[0, l_ms])
        if feasible != 'ok':
            return False
        return True

    def _interpolate_epoch(self, epoch, channel_set, bad_channels_idx,
                           no_neighbors=3):
        interp_epoch = epoch.copy()
        # Are all channels bad?
        if len(bad_channels_idx) == len(channel_set.channels):
            print('> Artifact rejection: Cannot interpolate because all '
                  'channels are bad.')
            return interp_epoch

        # For each bad channel
        bad_labels = []
        for i in bad_channels_idx:
            bad_labels.append(channel_set.channels[i]['label'])
        for i, bad_label in enumerate(bad_labels):
            if bad_label not in self.fitted['sorted_dist_ch']:
                raise Exception('Label %s is not present in the EEGChannelSet'
                                ' in which the model was fitted for!'
                                % bad_label)
            # Find the K labels of the nearest neighbors
            sorted_ch = self.fitted['sorted_dist_ch'][bad_label]
            interp_labels = []
            for ch in sorted_ch:
                interp_labels.append(ch["channel"]["label"])
                if len(interp_labels) == no_neighbors:
                    break

            # Interpolate using average
            interp_idxs = channel_set.get_cha_idx_from_labels(interp_labels)
            interp_epoch[:, bad_channels_idx[i]] = \
                np.mean(interp_epoch[:, np.array(interp_idxs)], axis=1)
        print('> Artifact rejection: interpolated %i channels' %
              len(bad_channels_idx))
        return interp_epoch

    def predict(self, times, signal, trial_idx, exp_data, sig_data):
        # Parameters
        len_epoch_ms = self.fitted['len_epoch_ms']
        len_epoch_sam = self.fitted['len_epoch_sam']
        fs = self.fitted['fs']
        exp_data.onsets = np.array(exp_data.onsets)

        # Assert filter bank
        if not isinstance(signal, list):
            signal = [signal]
        for seq_, seq_data_ in self.fitted['sequences'].items():
            if len(seq_data_) != len(signal):
                raise ValueError('[CircularShiftingClassifier] Cannot predict '
                                 'if the signal do not have the same number of '
                                 'filter banks than the fitted one!')

        # For each number of cycles
        pred_item_by_no_cycles = []
        no_cycles = np.max(exp_data.cycle_idx).astype(int) + 1
        for nc in range(no_cycles):
            # Identify what are the epochs that must be processed
            idx = (np.array(exp_data.trial_idx) == trial_idx) & \
                  (np.array(exp_data.cycle_idx) <= nc)

            # Raster latencies?
            raster_dict = None
            if self.correct_raster_latencies:
                possible_onsets_idx = np.where(
                    exp_data.raster_events['onset'] <
                    exp_data.onsets[idx][-1]
                )[0]
                if possible_onsets_idx.size > 0:
                    raster_dict = exp_data.raster_events['event'][
                        possible_onsets_idx[-1]]

            # For each fitted sequence
            pred_item = []
            for seq_, seq_data_ in self.fitted['sequences'].items():

                # For each possible filter bank
                f_corrs = []
                for filter_idx, filter_signal in enumerate(signal):

                    # Extract the epochs for that signal, trial and no. cycles
                    epochs = mds.get_epochs_of_events(
                        timestamps=times,
                        signal=filter_signal,
                        onsets=exp_data.onsets[idx],
                        fs=fs,
                        w_epoch_t=[0, len_epoch_ms],
                        w_baseline_t=None,
                        norm=None)
                    if len(epochs.shape) == 2:
                        # Create a dummy dimension if we have only one epoch
                        epochs = np.expand_dims(epochs, 0)

                    # Artifact rejection
                    if self.art_rej is not None:
                        epoch_std_by_channel = \
                            np.std(epochs[:, :len_epoch_sam, :], axis=1)
                        for i in range(epoch_std_by_channel.shape[0]):
                            discard_epoch = epoch_std_by_channel[i, :] > \
                                            seq_data_[filter_idx][
                                                'std_by_channel'] \
                                            * self.art_rej
                            if np.any(discard_epoch):
                                # TODO: precompute distance matrix before
                                epochs[i, :len_epoch_sam, :] = \
                                    self._interpolate_epoch(
                                        epoch=epochs[i, :len_epoch_sam, :],
                                        channel_set=sig_data.channel_set,
                                        bad_channels_idx=
                                        np.where(discard_epoch)[0],
                                        no_neighbors=3
                                    )

                    # Average the epochs
                    avg = np.mean(epochs[:, :len_epoch_sam, :], axis=0)

                    # CCA projection
                    x_ = seq_data_[filter_idx]['cca'].project(
                        avg, filter_idx=0, projection='Wy')

                    # Correlation coefficients between x_ and the templates
                    corrs = []
                    seqs = []
                    for shift_seq_, template_ in \
                            seq_data_[filter_idx]['templates'].items():
                        # Correct template using raster latencies
                        lat_s = 0
                        if raster_dict is not None:
                            if shift_seq_ in raster_dict:
                                lat_s = int(raster_dict[shift_seq_] * fs)
                        tem_ = np.roll(template_, lat_s)
                        temp_p = np.dot(tem_, x_) / np.sqrt(np.dot(np.dot(
                            tem_, tem_), np.dot(x_, x_)))
                        corrs.append(temp_p)
                        seqs.append(shift_seq_)
                    f_corrs.append(corrs)

                # Average the correlations between different filter banks
                corrs = np.mean(np.array(f_corrs), axis=0)
                seqs = np.array(seqs)

                # Sort the sequences by corrs' descending order
                sorted_idx = np.argsort(-corrs)
                sorted_corrs = corrs[sorted_idx]
                sorted_seqs = seqs[sorted_idx, :]

                # Identify the selected command
                sorted_cmds = get_items_by_sorted_sequences(
                    experiment=exp_data,
                    trial_idx=trial_idx,
                    sorted_seqs=sorted_seqs,
                    sorted_corrs=sorted_corrs
                )
                pred_item.append({
                    'sorted_cmds': sorted_cmds,
                    'fitted_sequence': seq_
                })

            # Store the predicted item
            pred_item_by_no_cycles.append(pred_item)

        return pred_item_by_no_cycles

    def predict_dataset(self, dataset: CVEPSpellerDataset,
                        show_progress_bar=True):
        # Error detection
        if not self.fitted:
            raise Exception(
                'Cannot predict if circular shifting templates and '
                'CCA projections are not fitted before! Aborting...')
        for rec in dataset.recordings:
            rec_sig = getattr(rec, dataset.biosignal_att_key)
            rec_exp = getattr(rec, dataset.experiment_att_key)
            if rec_sig.fs != self.fitted['fs']:
                raise ValueError('The sampling rate of this test recording '
                                 '(%.2f Hz) is not the same as for the fitted '
                                 'recordings! (%.2f Hz)' %
                                 (rec_sig.fs, self.fitted['fs']))
            if rec_exp.fps_resolution != self.fitted['fps_resolution']:
                raise ValueError('The refresh rate of this test recording '
                                 '(%.2f Hz) is not the same as for the fitted '
                                 'recordings! (%.2f Hz)' %
                                 (rec_exp.fps_resolution,
                                  self.fitted['fps_resolution']))

        if show_progress_bar:
            pbar = tqdm(total=len(dataset.recordings),
                        desc='Predicting dataset')

        # For each recording
        pred_items_by_no_cycles = []
        for rec in dataset.recordings:
            rec_sig = getattr(rec, dataset.biosignal_att_key)
            rec_exp = getattr(rec, dataset.experiment_att_key)

            # For each trial
            for t_idx in np.unique(rec_exp.trial_idx):
                decoding_by_no_cycles = \
                    self.predict(rec_sig.times, rec_sig.signal, t_idx,
                                 rec_exp, rec_sig)
                pred_items_by_no_cycles.append(decoding_by_no_cycles)
            if show_progress_bar:
                pbar.update(1)
        if show_progress_bar:
            pbar.close()
        return pred_items_by_no_cycles


class CircularShiftingEarlyStopping(components.ProcessingMethod):
    def __init__(self, **kwargs):
        """ Class constructor """
        super().__init__()

    def check_early_stop(self, corr_vector, std=3.0):
        """ Early stopping method based on normal distributions.

        Parameters
        --------------
        corr_vector: list() or 1D ndarray
            Vector that represents the sorted correlations for each of the
            possible commands, where corr_vector[0] must point to the most
            probable selected command.
        std: int
            Multiplier that determines if the selected command is an outlier
            of the normal distribution made up from the rest of correlations.
            Typical values are: 1 (outside 68% of data), 2 (outside 95% of
            data), and 3 (default, outside 99.7% of data).

        Returns
        --------------
        must_stop: bool
            True if it is possible to stop now, false otherwise.
        probs: 1D ndarray
            Current estimated probabilities of being selected (sorted).
        """
        corr_vector = np.array(corr_vector)
        threshold = np.mean(corr_vector[1:]) + std * np.std(corr_vector[1:])
        must_stop = corr_vector[0] > threshold
        probs = threshold - corr_vector
        probs = 1 - (probs / np.max(probs))
        return must_stop, probs


# ------------------------------- UTILS -------------------------------------- #
def get_unique_sequences_from_targets(experiment: CVEPSpellerData):
    """ Function that returns the unique sequences of all targets.
            return
        """
    sequences = dict()
    try:
        # todo: command_idx, unit_idx y demas lo tiene que hacer medusa y no unity

        for idx in range(len(experiment.command_idx)):
            # todo: revisar lo de los levels
            # Get the sequence used for the current command
            # l_ = int(experiment.level_idx[idx])
            # u_ = int(experiment.unit_idx[idx])
            # curr_command = experiment.paradigm_conf[m_][l_][u_][c_]
            m_ = int(experiment.matrix_idx[idx])
            c_ = int(experiment.command_idx[idx])
            curr_seq_ = experiment.commands_info[m_][str(c_)]['sequence']
            # Note: str(c_) is used because the previous json serialization
            #       interprets all dictionary keys as strings.

            # Add the command index to its associated sequence
            if tuple(curr_seq_) not in sequences:
                sequences[tuple(curr_seq_)] = [idx]
            else:
                sequences[tuple(curr_seq_)].append(idx)

        # todo: check that sequences are not shifted versions of themselves??
    except Exception as e:
        print(e)
    return sequences


def check_if_shifted(seq1, seq2):
    max_corr = np.max(np.correlate(seq1, seq1, 'same'))
    cross_corr = np.correlate(seq1, seq2, 'same')
    if np.max(cross_corr) == max_corr:
        print('WARNING: Two sequences are shifted versions of themselves!')
        return True
    else:
        return False


def get_items_by_sorted_sequences(experiment: CVEPSpellerData,
                                  trial_idx, sorted_seqs, sorted_corrs=None):
    # Find the first index of the trial to access matrix_idx, etc
    try:
        idx = list(experiment.trial_idx == trial_idx).index(True)
    except ValueError as e:
        raise ValueError('[get_items_by_sorted_sequences] Trial with idx %i not'
                         ' found in the experiment data! ' + str(e) % trial_idx)

    # Get the possible commands
    m_ = int(experiment.matrix_idx[idx])
    l_ = int(experiment.level_idx[idx])
    u_ = int(experiment.unit_idx[idx])
    possible_cmd = experiment.paradigm_conf[m_][l_][u_]

    # For each sequence in descending order of probability of being selected
    sorted_commands = list()
    for i in range(sorted_seqs.shape[0]):
        # For each possible command
        curr_comm_dict = dict()
        for cmd_id in possible_cmd:
            cmd_seq = experiment.commands_info[m_][str(cmd_id)]['sequence']
            if np.all(np.array(cmd_seq) == sorted_seqs[i, :]):
                # Found!
                curr_comm_dict['item'] = experiment.commands_info[m_][str(
                    cmd_id)]
                curr_comm_dict['label'] = curr_comm_dict['item']['label']
                curr_comm_dict['coords'] = [m_, l_, u_, int(cmd_id)]
                curr_comm_dict['correlation'] = None
                if sorted_corrs is not None:
                    curr_comm_dict['correlation'] = sorted_corrs[i]
                sorted_commands.append(curr_comm_dict)
                break
    return sorted_commands


def autocorr_zeropad(x):
    """ With zero padding, equivalent to np.correlate() """
    N = len(x)
    rxx = []
    x_lagged = np.concatenate((np.zeros((N - 1,)), x, np.zeros((N,))))
    for i in range(2 * N - 1):
        rxx.append(np.sum(x * x_lagged[i:i + N]))
    rxx = np.array(rxx)
    return rxx


def autocorr_circular(x):
    """ With circular shifts (periodic correlation) """
    N = len(x)
    rxx = []
    for i in range(-(N - 1), N):
        rxx.append(np.sum(x * np.roll(x, i)))
    rxx = np.array(rxx)
    return rxx


# ----------------------------- CODE GENERATORS ----------------------------- #
class LFSR:
    """ Computes a Linear-Feedback Shift Register (LFSR) sequence. """

    def __init__(self, polynomial, base=2, seed=None, center=False):
        """ Constructor of LFSR """
        self.polynomial = polynomial
        self.base = base
        self.seed = seed
        self.order = len(polynomial)
        self.N = base ** self.order - 1
        self.sequence = self.lfsr(polynomial, base, seed, center)

    @staticmethod
    def lfsr(polynomial, base=2, seed=None, center=False):
        """ This method implements a Linear-Feedback Shift Register (LFSR).

        IMPORTANT: maximal length sequences (m-sequences) can be only generated
        if the polynomial (taps) is primitive. I.e.:
            - the number of taps is even.
            - the set of taps is setwise co-prime (there must be no divisor
            other than 1 common to all taps).
        A list of primitive polynomials in function of the order m can be
        found here:
        https://en.wikipedia.org/wiki/Linear-feedback_shift_register

        NOTE: if the seed is composed by all zeros, the output sequence will be
        zeros.

        Parameters
        ----------
        polynomial: list
            Generator polynomial. E.g. (bias is specified for math convention
            but not used):
            "1 + x^5 + x^6" would be [0, 0, 0, 0, 1, 1].
            "1 + 2x + x^4" would be [2, 0, 0, 1]
        base : int
            (Optional, default: base = 2) Base of the sequence events that
            belongs to the Galois Field of the same base. By default, base=2,
            so only events of type {0,1} (or {-1,1} are part of the returned
            sequence.
        seed : list
            (Optional) Initial state. If not provided, the default state is a
            one-array with length equal to the order of the polynomial.
        center : bool
            (Optional, default = False) Determines if a centering over zero
            must be performed in the returned sequence (e.g., {0,1} -> {-1,1})

        Returns
        -------
        sequence : list
            LFSR m-sequence (maximum length is base^order-1), where order is the
            order of the polynomial.
        """
        # Defaults and error detection
        order = len(polynomial)
        if seed is None:
            seed = [1 for i in range(order)]
        if order > len(seed):
            raise Exception('[LSFR] The order of the polynom (%i) is higher '
                            'than the initial state length (%i)!' %
                            (order, len(seed)))

        # LFSR
        sequence = seed.copy()
        polynom = np.array(polynomial)
        while len(sequence) < base ** order - 1:
            new_bit = (np.matmul(polynom, np.array(sequence[:order]))) % base
            sequence.insert(0, new_bit)

        # Map the values to center around zero
        if center:
            if base == 2:
                sequence = np.array(sequence) * 2 - 1
            else:
                sequence = np.array(sequence) - np.floor(base / 2).astype(int)
        return sequence



# Built-in packages
import pickle, json

# External packages
from flask import request, Response, Blueprint
import bson

# Internal packages
from medusa.bci.remote import erp_spellers_api

erp_spellers_routing = Blueprint('bci_routing', 'medusa-server', template_folder='templates')


@erp_spellers_routing.route("/erp-speller/<paradigm>/train-model/", methods=['POST'])
def bci_erp_spellers_train_model(paradigm):
    try:
        # Get the info and files attached in the HTTP request
        info = json.loads(request.json)
        files = request_files_to_dict(request)

        # If there is no runs attached, abort
        if "run" not in files:
            raise Exception("No runs attached")
        bin_runs = files['run']

        # If there is no runs attached, abort
        if len(bin_runs) == 0:
            raise Exception("No runs attached")

        # Train models
        resp_content = erp_spellers_api.train_model(paradigm, info, bin_runs)

        # Response request
        bin_resp_content = bson.dumps(resp_content)
        resp = Response(bin_resp_content, status=200, mimetype='application/bson')

    except Exception as e:
        resp_content = {"error": str(e)}
        bin_resp_content = bson.dumps(resp_content)
        resp = Response(bin_resp_content, status=500, mimetype='application/bson')
    return resp


@erp_spellers_routing.route("/erp-spellers/<paradigm>/start-session/", methods=['POST'])
def start_rcp_session(paradigm):
    try:
        # Get the info and files attached in the HTTP request
        info = json.loads(request.json)
        files = request_files_to_dict(request)

        # ERP model is mandatory
        if 'erp_model' in files:
            erp_model = pickle.loads(files['erp_model'])
        else:
            raise Exception("No ERP model attached")

        # Control state model is not mandatory
        if 'cs_model' in files:
            cs_model = pickle.loads(files['erp_model'])
        else:
            cs_model = None

        # Create an ERPSession
        erp_session = erp_spellers_api.ERPSpellerSession(erp_model, cs_model,
                                                         session_timeout=1800,
                                                         ports=list(range(50001, 50007)))
        # Start the session
        erp_session.start()
        # Response content
        resp_content = {"port": erp_session.port}
        bin_resp_content = bson.dumps(resp_content)
        resp = Response(bin_resp_content, status=200, mimetype='application/bson')

    except Exception as e:
        resp_content = {"error": str(e)}
        bin_resp_content = bson.dumps(resp_content)
        resp = Response(bin_resp_content, status=500, mimetype='application/bson')
    return resp



"""
Created on Mon Mar 21 15:34:18 2022

@author: Diego Marcos-Martínez
"""
# Python modules
import argparse
import os
# Medusa modules
from medusa.meeg import EEGChannelSet

# External modules
import numpy as np
from vispy import scene, app
from vispy.io import imread, load_data_file, read_mesh
from vispy.scene.visuals import Mesh, Markers, Text, Graph
from vispy.scene import transforms
from vispy.visuals.filters import TextureFilter
from vispy.color import get_colormap
from medusa.meeg import EEGChannelSet


class TridimentionalBrain():
    def __init__(self, bg_color='black', text_color='white', translucent=None,
                 subplots=None, models=None, names=None):

        self.subplots = subplots
        self.bg_color = bg_color
        self.text_color = text_color
        self.translucent = translucent
        self.models = models
        self.names = names

        # Initialize attributes
        self.canvas = None
        self.cameras = None
        self.views = []
        self.grid = None
        self.standards = []
        self.sources = []
        self.brain_visuals = []
        self.markers = None
        self.labels_text = None
        self.connections_coords = None
        self.connections_values = None
        self.lines = None
        self.lines_cmap = None
        self.n_subplots = None

        self.__check_subplot()

        # Set canvas and view
        self.__set_canvas()

        # Set alpha
        self.__set_alpha()

        # Set brain models
        self.__set_models()

        # Set brain visuals
        self.__set_brain_visual()

        # Initializing methods
        self.__initialize_connections()

        self.canvas.show()

    def __check_subplot(self):
        """ This function ensures subplots parameter is correctly defined"""
        try:
            if self.subplots is not None:
                assert isinstance(self.subplots, tuple)
                assert len(self.subplots) == 2
                self.n_subplots = self.subplots[0] * self.subplots[1]
            else:
                self.n_subplots = 1
                self.subplots = (1, 1)
        except ValueError:
            print("Subplots must be two-dimensional tuple or None")

    def __set_alpha(self):
        try:
            if self.translucent is None:
                self.translucent = np.ones(self.n_subplots, dtype=bool)
            elif isinstance(self.translucent,bool):
                self.translucent = [self.translucent]
            else:
                assert len(self.translucent) == self.n_subplots
        except Exception as ex:
            print(ex)

    def __set_models(self):
        try:
            if self.models is None:
                self.models = ['B1']
                self.models = list(np.tile(self.models, self.n_subplots))
            else:
                assert len(self.models) == self.n_subplots
        except Exception as ex:
            print(ex)

    def __set_cameras(self):
        try:
            self.cameras = []
            for i in range(self.n_subplots):
                self.cameras.append('turntable')
        except Exception as ex:
            print(ex)

    def __set_canvas(self):
        try:
            self.canvas = scene.SceneCanvas(keys='interactive',
                                            bgcolor=self.bg_color,
                                            size=(850, 850), show=True,
                                            dpi=200)

            self.__set_cameras()

            if self.names is not None:
                assert len(self.names) == self.n_subplots
            self.grid = self.canvas.central_widget.add_grid(margin=10)
            self.grid.spacing = 0
            _view_idx = 0
            for row in range(self.subplots[0]):
                for col in range(self.subplots[1]):
                    if self.n_subplots is None:
                        view = self.grid.add_view(row=row, col=col)
                    else:
                        if self.names is not None:
                            assert isinstance(self.names[_view_idx], str)
                            title = scene.Label(self.names[_view_idx],
                                                color=self.text_color)
                            title.height_max = 60
                            self.grid.add_widget(title, row=row, col=col,
                                                 col_span=1)
                            view = self.grid.add_view(row=row + 1, col=col)
                        else:
                            view = self.grid.add_view(row=row, col=col)
                    view.camera = self.cameras[_view_idx]
                    self.views.append(view)
                    _view_idx += 1
        except Exception as ex:
            print(ex)

    def __set_brain_visual(self):
        try:
            parser = argparse.ArgumentParser()
            parser.add_argument('--shading', default='flat',
                                choices=['none', 'flat', 'smooth'],
                                help="shading mode")
            args, _ = parser.parse_known_args()
            shading = None if args.shading == 'none' else args.shading

            for i in range(self.n_subplots):
                path = os.path.join(os.path.dirname(__file__),
                                    'templates/{}.npz'.format(self.models[i]))
                model = np.load(path)
                vertices, faces = model.f.vertices, model.f.faces

                # This is temporary, just to adjust shape and coordinates
                vertices[:, 0] = 0.85 * vertices[:, 0] / np.max(vertices[:, 2])
                vertices[:, 1] = 0.85 * vertices[:, 1] / np.max(vertices[:, 2])
                vertices[:, 2] = 0.85 * vertices[:, 2] / np.max(vertices[:, 2])

                # Adapt the depth to the scale of the mesh to avoid rendering artefacts.
                self.views[i].camera.depth_value = 10 * (
                            vertices.max() - vertices.min())

                # Define transparency of each brain
                if self.translucent[i]:
                    alpha = 0.3
                else:
                    alpha = 1
                mesh = Mesh(vertices, faces, shading=shading,
                            color=(1, 1, 1, alpha))
                mesh.shading_filter.shininess = 1e+2
                self.brain_visuals.append(mesh)
                self.attach_headlight(mesh, self.views[i], self.canvas)
        except Exception as ex:
            print(ex)

    def add_brains(self):
        try:
            for view_idx in range(len(self.views)):
                self.views[view_idx].add(self.brain_visuals[view_idx])
        except Exception as ex:
            print(ex)

    def attach_headlight(self, mesh, view, canvas):
        """This function sets the initial light direction """
        light_dir = (1, 0, 1, 1)
        mesh.shading_filter.light_dir = light_dir[:3]
        initial_light_dir = view.camera.transform.imap(light_dir)

        @view.scene.transform.changed.connect
        def on_transform_change(event):
            """ This function calculates the illumination every
            time the brain is rotated"""
            transform = view.camera.transform
            mesh.shading_filter.light_dir = transform.map(initial_light_dir)[:3]

    def set_markers(self, locs, sub_plot):
        try:
            if self.markers is None:
                self.__init_markers()
            markers = Markers(light_color=self.text_color,size = 100)
            markers.set_data(locs)
            self.markers[sub_plot[0]][sub_plot[1]] = markers

            _view_idx = self.__calculate_subplot_idx(sub_plot)
            self.views[_view_idx].add(markers)
        except Exception as ex:
            print(ex)

    def set_labels(self, labels, locs, sub_plot):
        try:
            if self.labels_text is None:
                self.__init_labels()

            # Move the labels over the channels
            text_coord = locs.copy()
            text_coord[:, 2] += 0.1
            labels_text = Text(labels, color=self.text_color, bold=True,
                               pos=text_coord, font_size=50)
            self.labels_text[sub_plot[0]][sub_plot[1]] = labels_text

            _view_idx = self.__calculate_subplot_idx(sub_plot)
            self.views[_view_idx].add(labels_text)
        except Exception as ex:
            print(ex)

    def __set_conn_color(self, sub_plot, threshold, clim):
        try:
            # Mask under threshold values
            mask = np.ones(
                len(self.connections_values[sub_plot[0]][sub_plot[1]]),
                dtype=bool)
            mask[np.where(np.abs(self.connections_values[sub_plot[0]][
                                     sub_plot[1]]) < threshold)] = 0

            # Delete under threshold values and their connection coordinates
            self.connections_values[sub_plot[0]][sub_plot[1]] = \
                self.connections_values[sub_plot[0]][sub_plot[1]][mask]
            self.connections_coords_mutable[sub_plot[0]][sub_plot[1]] = \
                self.connections_coords_invariant[sub_plot[0]][sub_plot[1]][
                    mask]

            # Duplicate the matrix of values so that each point of the
            # connection has the same value.
            self.connections_values[sub_plot[0]][sub_plot[1]] = \
                np.tile(self.connections_values[sub_plot[0]][sub_plot[1]],
                        (2, 1)).reshape(2 * len(self.connections_values[
                                                    sub_plot[0]][sub_plot[1]]),
                                        order='F')

            if clim is None:
                clim = []
                clim.append(np.round(np.min(self.connections_values[sub_plot[0]]
                                        [sub_plot[1]]),decimals=2))
                clim.append(np.round(np.max(self.connections_values[sub_plot[0]]
                                   [sub_plot[1]]),decimals = 2))

            color = self.lines_cmap.map((self.connections_values[sub_plot[0]]
                                        [sub_plot[1]] - clim[0])/(clim[1] - clim[0]))
            return color, clim
        except Exception as ex:
            print(ex)

    def set_connections(self, adj_mat, locs, sub_plot=None, threshold=0.5,
                        plot_markers=True, labels=None, plot_labels=False,
                        cmap='seismic', clim = None, cbar = False):
        try:
            if sub_plot is None and self.n_subplots == 1:
                sub_plot = (0, 0)
            assert isinstance(sub_plot, tuple)
            assert sub_plot[0] <= self.subplots[0] and sub_plot[1] <= \
                   self.subplots[1]
            if clim is not None:
                assert  isinstance(clim,list)

            if plot_labels:
                if labels is None:
                    print("Labels could not been added because labels parameter"
                          "is None")
                else:
                    assert all(isinstance(elem, str) for elem in labels)
                    self.set_labels(labels, locs, sub_plot)

            if plot_markers:
                self.set_markers(locs, sub_plot)


            # Extract connectivity values
            self.connections_values[sub_plot[0]][
                sub_plot[1]] = self.__extract_conn_values(adj_mat)

            # Set connections matrix
            self.connections_coords_invariant[sub_plot[0]][
                sub_plot[1]] = self.__set_connections_coords(locs,
                                                             self.connections_values
                                                             [sub_plot[0]][
                                                                 sub_plot[
                                                                     1]])

            # Get color map and color connections
            self.lines_cmap = get_colormap(cmap)
            color, clim = self.__set_conn_color(sub_plot, threshold, clim)


            _view_idx = self.__calculate_subplot_idx(sub_plot)
            self.lines[sub_plot[0]][sub_plot[1]] = scene.Line(antialias=True,
                                                              parent=self.views[
                                                                  _view_idx].scene,
                                                              )
            self.lines[sub_plot[0]][sub_plot[1]].set_data(
                pos=self.connections_coords_mutable[sub_plot[0]][sub_plot[1]],
                color=color, width=4)

            # Plot color bar (In process)
            if cbar:
                cbar_widget = scene.ColorBarWidget(clim=clim, cmap=cmap,
                                                   orientation='bottom',)

                self.grid.add_widget(cbar_widget, row=sub_plot[0]+2,
                                     col=sub_plot[1])

        except Exception as ex:
            print(ex)

    def update_connections(self, adj_mat, sub_plot, threshold):
        try:
            if sub_plot is None:
                sub_plot = (0, 0)
            if threshold is None:
                threshold = 0.5
            # Extract connectivity values
            self.connections_values[sub_plot[0]][
                sub_plot[1]] = self.__extract_conn_values(adj_mat)
            color = self.__set_conn_color(sub_plot, threshold)
            _view_idx = self.__calculate_subplot_idx(sub_plot)

            self.lines[sub_plot[0]][sub_plot[1]].parent = None

            self.lines[sub_plot[0]][sub_plot[1]] = scene.Line(antialias=True,
                                                              parent=self.views[
                                                                  _view_idx].scene)
            self.lines[sub_plot[0]][sub_plot[1]].set_data(
                pos=self.connections_coords_mutable[sub_plot[0]][sub_plot[1]],
                color=color, width=4, )
            self.canvas.update()
        except Exception as ex:
            print(ex)

    def __calculate_subplot_idx(self, sub_plot):
        try:
            _view_idx = sub_plot[1] + self.subplots[1] * sub_plot[0]
            return _view_idx
        except Exception as ex:
            print(ex)

    def __initialize_connections(self):
        try:
            subplots = self.subplots
            if self.n_subplots is None:
                subplots = (1, 1)
            self.connections_coords_invariant = np.empty(
                shape=subplots + (0,)).tolist()
            self.connections_coords_mutable = np.empty(
                shape=subplots + (0,)).tolist()
            self.connections_values = np.empty(shape=subplots + (0,)).tolist()
            self.lines = np.empty(shape=subplots + (0,)).tolist()
        except Exception as ex:
            print(ex)

    def __init_markers(self):
        try:
            subplots = self.subplots
            self.markers = np.empty(shape=subplots + (0,)).tolist()
        except Exception as ex:
            print(ex)

    def __init_labels(self):
        try:
            subplots = self.subplots
            self.labels_text = np.empty(shape=subplots + (0,)).tolist()
        except Exception as ex:
            print(ex)

    @staticmethod
    def __extract_conn_values(adj_mat):
        try:
            values = np.triu(adj_mat, 1)
            values = values[np.where(values != 0)]
            return values
        except Exception as ex:
            print(ex)

    @staticmethod
    def __set_connections_coords(locs, connections_values):
        try:
            connections_coords = np.empty((len(connections_values), 2, 3))
            value_idx = 0
            for i in range(len(locs)):
                for j in range(i + 1, len(locs)):
                    connections_coords[value_idx, 0, 0] = locs[i, 0]
                    connections_coords[value_idx, 0, 1] = locs[i, 1]
                    connections_coords[value_idx, 0, 2] = locs[i, 2]
                    connections_coords[value_idx, 1, 0] = locs[j, 0]
                    connections_coords[value_idx, 1, 1] = locs[j, 1]
                    connections_coords[value_idx, 1, 2] = locs[j, 2]
                    value_idx += 1
            return connections_coords
        except Exception as ex:
            print(ex)




if __name__ == '__main__':
    from vispy.app import use_app

    app = use_app("pyqt5")
    app.create()
    # Set canvas
    triplot = TridimentionalBrain( bg_color='white',
                                  text_color='black',translucent=[True,True],subplots=(1,2), names=['Prueba1','Prueba2'])

    # Define channel set and its coord
    channel_set = EEGChannelSet(dim='3D', coord_system='cartesian')
    channel_set.set_standard_montage(montage='10-20')
    channel_coord = np.zeros((len(channel_set.channels), 3))
    for ch_idx, channel in enumerate(channel_set.channels):
        channel_coord[ch_idx, 0] = 0.8 * (channel['x'])
        channel_coord[ch_idx, 1] = 0.9 * (channel['y'] - 0.2)
        channel_coord[ch_idx, 2] = 0.85 * (channel['z'])
    adj_mat = np.random.randn(len(channel_set.channels),
                              len(channel_set.channels))
    adj_mat = 2 * adj_mat - 1
    triplot.set_connections(adj_mat, channel_coord, threshold=[0.6], plot_labels=True,
                            labels=channel_set.l_cha, plot_markers=True,cmap='Spectral',sub_plot=(0,0))
    triplot.set_connections(adj_mat, channel_coord, threshold=[0.6], plot_labels=False,
                            labels=channel_set.l_cha, plot_markers=True,cmap='Spectral',sub_plot=(0,1))
    triplot.add_brains()
    triplot.canvas.show()
    app.run()



"""
In this module you will find useful functions to apply optimized phase-based
connectivity metrics. Enjoy!

@authors: Víctor Rodríguez-González and Diego Marcos-Martínez
"""

# Built-in imports
import warnings

# External imports
import scipy.signal as sp_signal
import numpy as np
import tensorflow as tf

# Medusa imports
from medusa import transforms
from medusa.utils import check_dimensions


def __phase_connectivity_cpu(data, measure=None):
    """ This method implements three phase-based connectivity parameters using
    CPU: PLV, PLI, and wPLI.

    Parameters
    ----------
    data : numpy 2D matrix
        MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels].
    measure: str or None
        Key of the phase connectivity measure to calculate: "plv", "pli" or
        "wpli". If None, the three phase connectivity measures are calculated.

    Returns
    -------
    plv : numpy 3D square matrix
        plv-based connectivity matrix. [n_epochs, n_channels x n_channels].
    pli : numpy 3D square matrix
        pli-based connectivity matrix. [n_epochs, n_channels x n_channels].
    wpli : numpy 3D square matrix
        wpli-based connectivity matrix. [n_epochs, n_channels x n_channels].

    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    # Variable initialization
    n_epochs = data.shape[0]
    n_samples = data.shape[1]
    n_chan = data.shape[2]

    # Connectivity computation
    phase_data = np.angle(transforms.hilbert(data))
    phase_data = np.ascontiguousarray(phase_data)
    angles_1 = np.reshape(np.tile(phase_data, (1, n_chan, 1)),
                          (n_epochs, n_samples, n_chan * n_chan),
                          order='F')
    angles_2 = np.tile(phase_data, (1, 1, n_chan))

    if measure == 'pli':
        return __pli(angles_1, angles_2, n_epochs, n_chan, 'CPU')
    elif measure == 'wpli':
        return __wpli(angles_1, angles_2, n_epochs, n_chan, 'CPU')
    elif measure == 'plv':
        return __plv(angles_1, angles_2, n_epochs, n_chan, 'CPU')
    else:
        plv = __plv(angles_1, angles_2, n_epochs, n_chan, 'CPU')
        pli = __pli(angles_1, angles_2, n_epochs, n_chan, 'CPU')
        wpli = __wpli(angles_1, angles_2, n_epochs, n_chan, 'CPU')
        return plv, pli, wpli


def __phase_connectivity_gpu(data, measure=None):
    """ This method implements three phase-based connectivity parameters using
    GPU: PLV, PLI, and wPLI.

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels] and
        [n_samples, n_channels].
    measure: str or None
        Key of the phase connectivity measure to calculate: "plv", "pli" or
        "wpli". If None, the three phase connectivity measures are calculated.

    Returns
    -------
    plv : numpy 3D square matrix
        plv-based connectivity matrix. [n_epochs, n_channels, n_channels].
    pli : numpy 3D square matrix
        pli-based connectivity matrix. [n_epochs, n_channels, n_channels].
    wpli : numpy 3D square matrix
        wpli-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    import tensorflow as tf
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    # Variable initialization
    n_epochs = data.shape[0]
    n_samples = data.shape[1]
    n_chan = data.shape[2]

    # Connectivity computation
    phase_data = tf.math.angle(transforms.hilbert(data))

    angles_1 = tf.transpose(tf.reshape(
        tf.transpose(tf.tile(phase_data, (1, n_chan, 1)),
                     perm=[0, 2, 1]),
        (n_epochs, n_chan * n_chan, n_samples)), perm=[0, 2, 1])

    angles_2 = tf.tile(phase_data, (1, 1, n_chan))

    if measure == 'pli':
        return __pli(angles_1, angles_2, n_epochs, n_chan, 'GPU')
    elif measure == 'wpli':
        return __wpli(angles_1, angles_2, n_epochs, n_chan, 'GPU')
    elif measure == 'plv':
        return __plv(angles_1, angles_2, n_epochs, n_samples, n_chan, 'GPU')
    else:
        plv = __plv(angles_1, angles_2, n_epochs, n_samples, n_chan, 'GPU')
        pli = __pli(angles_1, angles_2, n_epochs, n_chan, 'GPU')
        wpli = __wpli(angles_1, angles_2, n_epochs, n_chan, 'GPU')
        return plv, pli, wpli


def __plv(angles_1, angles_2, n_epochs, n_samples, n_chan, ctype='CPU'):
    """ This auxiliary method implements both CPU and GPU PLV calculations"""
    if ctype == 'CPU':
        plv_vector = np.divide(
            abs(np.sum(np.exp(1j * (angles_1 - angles_2)), axis=1)),
            n_samples)
        plv = np.reshape(plv_vector, (n_epochs, n_chan, n_chan), order='F')
    elif ctype == 'GPU':
        plv_vector = tf.math.divide(
            tf.math.abs(
                tf.math.reduce_sum(
                    tf.math.exp(
                        tf.math.scalar_mul(
                            1j,
                            tf.cast(
                                tf.math.subtract(angles_1, angles_2),
                                'complex64'))),
                    axis=1)),
            n_samples)
        plv = tf.reshape(plv_vector, (n_epochs, n_chan, n_chan))
    return plv


def __pli(angles_1, angles_2, n_epochs, n_chan, ctype='CPU'):
    """ This auxiliary method implements both CPU and GPU PLI calculations"""
    if ctype == 'CPU':
        pli_vector = abs(np.mean(np.sign(np.sin(angles_1 - angles_2)), axis=1))
        pli = np.reshape(pli_vector, (n_epochs, n_chan, n_chan), order='F')
    elif ctype == 'GPU':
        pli_vector = tf.math.abs(
            tf.math.reduce_mean(
                tf.math.sign(
                    tf.math.sin(tf.math.subtract(angles_1, angles_2))),
                axis=1))
        pli = tf.reshape(pli_vector, (n_epochs, n_chan, n_chan))
    return pli


def __wpli(angles_1, angles_2, n_epochs, n_chan, ctype='CPU'):
    """ This auxiliary method implements both CPU and GPU wPLI calculations"""
    if ctype == 'CPU':
        imz = np.sin(angles_1 - angles_2)
        with np.errstate(divide='ignore', invalid='ignore'):
            wpli_vector = np.divide(
                abs(np.mean(np.multiply(abs(imz), np.sign(imz)), axis=1)),
                np.mean(abs(imz), axis=1)
            )
        wpli = np.reshape(wpli_vector, (n_epochs, n_chan, n_chan), order='F')
    elif ctype == 'GPU':
        imz = tf.math.sin(tf.math.subtract(angles_1, angles_2))
        wpli_vector = tf.math.divide(
            tf.math.abs(tf.math.reduce_mean(
                tf.math.multiply(
                    tf.math.abs(imz),
                    tf.math.sign(imz)),
                axis=1)),
            tf.math.reduce_mean(tf.math.abs(imz), axis=1))
        wpli = tf.reshape(wpli_vector, (n_epochs, n_chan, n_chan))

    wpli = tf.linalg.set_diag(wpli, np.zeros((wpli.shape[0], wpli.shape[1])))
    return wpli


def phase_connectivity(data, measure=None):
    """ This method implements three phase-based connectivity parameters: PLV,
    PLI, and wPLI.

    REFERENCES:
    PLV: Mormann, F., Lehnertz, K., David, P., & Elger, C. E.
    (2000). Mean phase coherence as a measure for phase synchronization and its
    application to the EEG of epilepsy patients. Physica D: Nonlinear Phenomena,
    144(3-4), 358-369.

    PLI: Nolte, G., Bai, O., Wheaton, L., Mari, Z., Vorbach, S., & Hallett, M.
    (2004). Identifying true brain interaction from EEG data using the imaginary
    part of coherency. Clinical neurophysiology, 115(10), 2292-2307.

    wPLI: Vinck, M., Oostenveld, R., Van Wingerden, M., Battaglia, F., &
    Pennartz, C. M. (2011). An improved index of phase-synchronization for
    electrophysiological data in the presence of volume-conduction, noise and
    sample-size bias. Neuroimage, 55(4), 1548-1565.

    NOTE: PLV is sensitive to volume conduction effects

    Parameters
    ----------
    data : numpy matrix
        MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels],
        [n_samples, n_channels].
    measure: str or None
        Key of the phase connectivity measure to calculate: "plv", "pli" or
        "wpli". If None, the three phase connectivity measures are calculated.
    Returns
    -------
    plv : numpy 3D square matrix
        plv-based connectivity matrix. [n_epochs, n_channels, n_channels].
    pli : numpy 3D square matrix
        pli-based connectivity matrix. [n_epochs, n_channels, n_channels].
    wpli : numpy 3D square matrix
        wpli-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    from medusa import tensorflow_integration
    # Error check
    if not np.issubdtype(data.dtype, np.number):
        raise ValueError('data matrix contains non-numeric values')
    if measure is not None and (measure != 'plv' and measure != 'pli' and measure
                                != 'wpli'):
        raise ValueError('Unknown measure key. Available are: "plv", "pli" or'
                         '"wpli".')

    if tensorflow_integration.check_tf_config(autoconfig=True):
        if measure is None:
            plv, pli, wpli, = __phase_connectivity_gpu(data, measure)
            return np.asarray(plv), np.asarray(pli), np.asarray(wpli)
        else:
            ph_cnn_measure = __phase_connectivity_gpu(data, measure)
            return np.asarray(ph_cnn_measure)
    else:
        if measure is None:
            plv, pli, wpli, = __phase_connectivity_cpu(data, measure)
            return np.asarray(plv), np.asarray(pli), np.asarray(wpli)
        else:
            ph_cnn_measure = __phase_connectivity_cpu(data, measure)
            return np.asarray(ph_cnn_measure)



import numpy as np
import os, json, pickle, datetime
import matplotlib.pyplot as plt
from sklearn.model_selection import ParameterGrid
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF
from scipy.stats import norm
import warnings


class Grinder:
    def __init__(self, **hyparams):
        # Make hyparams dictionary
        opt_params = dict()
        for key, value in hyparams.items():
            value = [value] if not isinstance(value, list) else value
            opt_params[key] = np.arange(len(value))
        
        # Make grid
        self.hyparams = hyparams
        self.opt_params = opt_params
        self.hyparams_grid = list(ParameterGrid(hyparams))
        self.opt_grid = list(ParameterGrid(opt_params))
    
    def get_hyparams(self, idx):
        return self.hyparams_grid[idx]


class Optimizer:

    def __init__(self, obj, func, grinder, args=None, previous_hist=None):
        
        # Supported values
        self.supported_obj = ('minimize', 'maximize')
        self.supported_approach = ('grid', 'random', 'bayesian')
        
        # Check errors
        if not obj in self.supported_obj:
            raise ValueError("Parameter 'obj' must be one of these values: " +
                             str(self.supported_obj))
        
        if (args is not None) and (type(args)!=dict):
                raise Exception("Argument 'args' must be None or type dict")
                
        if (previous_hist is not None) and (type(previous_hist)!=dict):
                raise Exception("Argument 'previous_hist' must be None or type "
                                "dict")
        
        # Initialize
        self.obj = obj
        self.func = func
        self.args = args
        self.grinder = grinder
        
        # Initialize history and the optim directory
        if previous_hist is None:
            # Initialize directory
            dir_name = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")        
            self.optim_dir = os.getcwd() + '/optimization/optim_' + \
                             dir_name + '\\'
            os.makedirs(self.optim_dir)
            
            # Instantiate useful variables
            self.history = dict()
            self.history["grid_idx"] = list()
            self.history["score"] = list()
            self.history["info"] = list()
            self.history["optim_dir"] = self.optim_dir
        else:
            warnings.warn("Loading a previous history only makes sense if "
                          "arguments 'obj', 'func', 'grinder', 'args' and have "
                          "not changed.")
            self.history = previous_hist
            self.optim_dir = self.history["optim_dir"]
            try:
                os.makedirs(self.history["optim_dir"])
            except FileExistsError as e:
                pass
        
        # Bayesian optimization variables
        self.bayes_feat = None
        self.bayes_model = None
        self.bayes_n_samp = None
        self.rbf_length_scale = None
        self.bayes_parameters = False
        
        # Save the grinder
        with open(self.optim_dir + 'grinder.pkl', 'wb') as file:
            pickle.dump(self.grinder, file)
    
    def __pick_params(self, approach):
        if approach == 'grid':
            idx = self.history["grid_idx"][-1]+1 \
                if len(self.history["grid_idx"]) > 0 else 0
        elif approach == 'random':
            indexes = np.arange(len(self.grinder.opt_grid))
            pickable = np.delete(indexes, self.history["grid_idx"])
            np.random.shuffle(pickable)
            idx = self.__pick_new_indexes_randomly(1, repetition=False)[0]
        elif approach == 'bayesian':
            # Pick new random samples
            samp = self.__pick_new_indexes_randomly(
                int(self.bayes_n_samp), repetition=False
            )
            feat = self.__get_bayes_feat(samp)
            mu, std = self.bayes_model.predict(feat, return_std=True)
            # Get best feat until now
            with warnings.catch_warnings():
                # ignore generated warnings
                warnings.simplefilter("ignore")
                # Update best model
                if self.obj == 'maximize':
                    best = np.max(self.bayes_model.predict(self.bayes_feat))
                else:
                    best = np.min(self.bayes_model.predict(self.bayes_feat))
            # Calculate the probability of improvement or expected improvement
            probs = norm.cdf((np.squeeze(mu) - best) / (std+1e-9))
            if self.obj == 'maximize':
                idx = samp[np.argmax(probs)]
            else:
                idx = samp[np.argmin(probs)]
        else:
            pass
        hyparams = self.grinder.get_hyparams(idx)
        # Return
        return int(idx), hyparams
    
    def __pick_new_indexes_randomly(self, n, repetition=False):
        indexes = np.arange(len(self.grinder.opt_grid))
        pickable = np.delete(indexes, self.history["grid_idx"]) \
            if repetition is False else indexes
        np.random.shuffle(pickable)
        return pickable[0:n]
    
    def __get_bayes_feat(self, indexes):
        data = np.empty((len(indexes), len(self.grinder.opt_grid[0])))
        i = 0
        for idx in indexes:
            data[i, :] = [value for key, value in
                          self.grinder.opt_grid[idx].items()]
            i += 1
        return data
    
    def set_bayes_opt_params(self, n_samp, rbf_length_scale=1):
        self.bayes_n_samp = n_samp
        self.rbf_length_scale = rbf_length_scale
        self.bayes_parameters = True
    
    def optimize(self, max_iter, approach, save_history=True):
        # Check errors
        if max_iter > len(self.grinder.opt_grid):
            raise ValueError("Parameter 'max_iter' must be <= than the grid "
                             "size:" + str(len(self.grinder.opt_grid)))
            
        if not approach in self.supported_approach:
            raise ValueError("Parameter 'approach' must be one of these "
                             "values: " + str(self.supported_obj))
            
        if approach == 'bayesian':
            if len(self.history["score"]) == 0:
                raise Exception("No previous history for bayesian "
                                "optimization. Initialize with a random "
                                "optimization")
            if not self.bayes_parameters:
                raise Exception("You must set the bayesian optimization "
                                "parameters first. Call function "
                                "set_bayes_opt_params.")
                
        # Initialize bayesian model
        if approach == 'bayesian':
            if self.bayes_feat is None:
                self.bayes_feat = self.__get_bayes_feat(self.history["grid_idx"])
            if self.bayes_feat.shape[0] != len(self.history["grid_idx"]):
                self.bayes_feat = self.__get_bayes_feat(self.history["grid_idx"])
            # Define gaussian model
            rbf_kernel = RBF(length_scale=self.rbf_length_scale)
            self.bayes_model = GaussianProcessRegressor(kernel=rbf_kernel, normalize_y=True)
            # Fit model
            self.bayes_model.fit(self.bayes_feat, self.history["score"])
            
        print("\nOPTIMIZATION (" + approach + ")\n")
        
        # Start optimization
        for i in range(max_iter):
            print("===========================================================")
            print("ITERATION " + str(i+1) + "/" + str(max_iter) + "...\n")
            
            # Get params
            idx, hyparams = self.__pick_params(approach)
            
            # Evaluate function
            score, info = self.func(hyparams=hyparams, args=self.args)
            if (info is not None) and (type(info)!=dict):
                raise Exception("History must be None or type dict")
                
            # Update history
            self.history["grid_idx"].append(idx)
            self.history["score"].append(score)
            self.history["info"].append(info)

            # Update best model
            if self.obj == 'maximize':
                best_score = np.max(self.history["score"])
                best_idx = self.history["grid_idx"][np.argmax(
                    self.history["score"]
                )]
            else:
                best_score = np.min(self.history["score"])
                best_idx = self.history["grid_idx"][np.argmin(
                    self.history["score"]
                )]

            # Print info
            print()
            if approach == 'bayesian':
                pred = self.bayes_model.predict(self.__get_bayes_feat([idx]))[0]
                print("Bayes estim\t= %.5f" % (pred) + ", grid_idx = " +
                      str(idx))
            print("Score \t\t= %.5f" % (score) + ", grid_idx = " +
                  str(idx))
            print("Best score \t= %.5f" % (best_score) + ", grid_idx = " +
                  str(best_idx))
            print("===========================================================")
            print()
                
            # Update bayesian model
            if approach == 'bayesian':
                self.bayes_feat = np.concatenate((self.bayes_feat,
                                                  self.__get_bayes_feat([idx])),
                                                 axis=0)
                self.bayes_model.fit(self.bayes_feat, self.history["score"])
                
            # Save history
            if save_history:
                with open(self.optim_dir + 'history.pkl', 'wb') as file:
                    pickle.dump(self.history, file)
                # Save json history (human-readable)
                with open(self.optim_dir + 'history.json', 'w') as file:
                    file.write(json.dumps(self.history, indent=4))
        
        # Return
        return best_idx, best_score, self.grinder.get_hyparams(best_idx)
        
    def get_best_hyparams(self):
        # Get best model
        if len(self.history["score"]) > 0:
            if self.obj == 'maximize':
                best_score = np.max(self.history["score"])
                best_idx = self.history["grid_idx"][np.argmax(
                    self.history["score"]
                )]
            else:
                best_score = np.min(self.history["score"])
                best_idx = self.history["grid_idx"][np.argmin(
                    self.history["score"]
                )]
            
            return best_score, best_idx, self.grinder.get_hyparams(best_idx)
        else:
            raise Exception("The optimization has not started")
        





import numpy as np

def __distance_inv_wei(L):
    """
    Inverse distance matrix. Dijkstra's algorithm. Rubinov & Sporns 2010

    Parameters
    ----------
    L : numpy 2D matrix
        Connection-length matrix THIS IS NOT WEIGHTS MATRIX

    Returns
    -------
    D : numpy 2D matrix
        Distance matrix.
    B : numpy 2D matrix
        Number of edges matrix. 
    """
    n = L.shape[0]       
    D = np.ones((n,n)) * np.inf # Distance matrix
    np.fill_diagonal(D,0)
    
    for u in range(0,n):
        S = np.ones((1,n),dtype=np.int32) # Distance permanence placeholder
        L1 = np.copy(L)
        V = [[u]]
        while 1:
            S[0,V] = 0 # Distance u>V is now permanent
            L1[:,V] = 0 # Remove in-edges
            for v in V:
                T = np.argwhere(L1[v,:]) # Neighbours of shortest nodes
                D[u,T[:,1]] = np.min(np.concatenate((D[u,T[:,1]][:,np.newaxis], (D[u,v]+L1[v,T[:,1]])[:,np.newaxis]),axis=1),axis=1)
                
            if (D[u,S[0,:].astype(bool)] == 0).all(): # Empty = all nodes reached, Inf = some nodes cannot be reached
                break
            
            minD = np.min(D[u,S[0,:].astype(bool)])

            if minD == np.inf: # Empty = all nodes reached, Inf = some nodes cannot be reached
                break
        
            V = np.argwhere(D[u,:]==minD)

    np.fill_diagonal(D,np.inf)    
    D = 1/D
    
    return D   

def efficiency(W):
    """
    Calculates the graph efficiency. 
    Globally is the mean of the inverse shortests path length, and is 
    inversely related to path length. 
    Nodally/Locally is the same as globally but calculated on the neighbourhood of the
    node, and is related with clustering coefficient. 
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    global_eff : numpy array
        Global efficiency.   
    nodal_eff : numpy array
        Nodal/Local efficiency.

    """
    
    N = W.shape[0] # Number of graph nodes
    W1 = np.copy(W) # Graph copy
    A = np.where(W != 0,1,0) # Connections matrix
    idx = np.copy(A)
    W1 = np.where(idx,1/W1,0) # Distance matrix (inverse of the non-zero weights)
    
    # Global efficiency
    e = __distance_inv_wei(W1)
    global_eff = np.sum(np.squeeze(np.asarray(e))) / (N**2 - N)
    
    # Nodal/Local efficiency - Based on Rubinov's BCT toolbox
    nodal_eff = np.zeros((N,1))
    for u in range(0,N):
        V = np.argwhere(np.logical_or(A[u,:],A.T[:,u])) # Neighbours
        sw = W[u,V]**(1/3) + W.T[u,V]**(1/3) # Symmetrized weights vector
        W1_tmp = np.copy(W1[np.squeeze(V),:])
        e = __distance_inv_wei(W1_tmp[:,np.squeeze(V)]) # Inverse distance matrix
        se = e**(1/3) + e.T**(1/3) # Symmetrized inverse distance matrix
        num = np.sum(np.sum((sw.T*sw)*se))/2 # Numerator

        if num != 0:
            sa = A[u,V] + A.T[u,V] # Symmetrized adjacency vector
            den = np.sum(sa)**2 - np.sum(sa**2) # Denominator
            nodal_eff[u] = num / den # Nodal/Local efficiency
            
    return global_eff,np.squeeze(nodal_eff)



"""Created on Monday March 15 19:27:14 2021

In this module, you will find all functions and data structures related to
EEG and MEG signals. Enjoy!

References:
    https://stefanappelhoff.com/eeg_positions
    https://github.com/sappelhoff/eeg_positions
    https://mne.tools/dev/auto_tutorials/intro/40_sensor_locations.html
@author: Eduardo Santamaría-Vázquez
"""

# Built-in imports
import math

# External imports
import warnings

import scipy.io
import numpy as np

# Medusa imports
from medusa import components
from medusa import meeg


class MEEGChannel(components.SerializableComponent):
    """This class implements a M/EEG channel.
    """

    # TODO: This class is till not being used for compatibility reasons, but it
    #  should be introduced in the next major update of medusa kernel to remove
    #  channel dictionaries and simplify the management of MEEG signals

    def __init__(self, label, coordinates=None, reference=None):
        """Constructor for class MEEGChannel.

        Parameters
        ----------
        label: str
            Label of the channel.
        coordinates: dict, optional
            Dict with the coordinates of the electrode. It is strongly
            recommended to set this optional parameter in order to use advanced
            features of MEDUSA (e.g., topographic plots).
        reference: MEEGChannel, optional
            Use only for bipolar montages to define the reference electrode.
        """
        # Check errors
        if reference is not None:
            if not isinstance(reference, MEEGChannel):
                raise ValueError(
                    'Parameter reference must be of type MEEGChannel or None')
        if coordinates is None:
            warnings.warn('Channel coordinates are used by some advanced '
                          'features of MEDUSA (e.g., topographic plots). '
                          'Please consider setting the coordinates or use '
                          'function MEEGChannel.from_standard_set to load '
                          'standard coordinates.')
        # Set attributes
        self.label = label
        self.coordinates = coordinates
        self.reference = reference

    def to_serializable_obj(self):
        pass

    @classmethod
    def from_standard_set(cls, label, standard='10-5'):
        pass

    @classmethod
    def from_serializable_obj(cls, data):
        pass


class EEGChannelSet(components.SerializableComponent):
    """Class to represent an EEG montage with ordered channels in specific
    coordinates. It also provides functionality to load channels from EEG
    standards 10-20, 10-10 and 10-5 directly from the labels.
    """

    def __init__(self, reference_method='common', dim='2D',
                 coord_system='spherical'):
        """Constructor of class EEGChannelSet

        Parameters
        ----------
        reference_method: str {'common'|'average'|'bipolar'}
            Reference method. Recordings with common reference are referenced
            to the same channel (e.g., ear lobe, mastoid). Recordings with
            average reference are referenced to the average of all or several
            channels. Finally, bipolar reference is the subtraction of 2
            channels.
        dim: str {'2D'|'3D'}
            Dimensions of the coordinates plane.
        coord_system: str {'cartesian'|'spherical'}
            Coordinates system. Take into account that, if dim = '2D' spherical
            refers to polar coordinates
        """

        # Check errors
        if reference_method not in ('common', 'average', 'bipolar'):
            raise ValueError('Unknown reference method %s' % reference_method)
        if dim not in ('2D', '3D'):
            raise ValueError('Unknown number of dimensions %s' % dim)
        if coord_system not in ('cartesian', 'spherical'):
            raise ValueError('Unknown coordinates system %s' % coord_system)

        # Set attributes
        self.reference_method = reference_method
        self.dim = dim
        self.coord_system = coord_system
        self.channels = None
        self.n_cha = None
        self.l_cha = None
        self.montage = None
        self.ground = None
        self.allow_unlocated_channels = False

    def set_ground(self, ground):
        """Sets the ground of the montage

        Parameters
        ----------
        ground: dict
            Dict with the ground data. The easiest way to calculate it is using
            function get_standard_channel_data_from_label. Keys:
                - label: channel label
                - coordinates: depends on the coordinate system and dimensions:
                        - Dim = 2D:
                            - Cartesian coordinates. Ex: {'x': 0.5, 'y': 0}
                            - Spherical coordinates Ex: {'r': 0.5,
                                'theta': np.pi/2}
                        - Dim = 3D:
                            - Cartesian coordinates. Ex: {'x': 0.5,
                                'y': 0, 'z': 0.8}
                            - Spherical coordinates. Ex: {'r': 0.5,
                                'theta': np.pi/2, 'phi': np.pi/4}
        """
        self.ground = ground

    def add_channel(self, channel, reference):
        """Function to add a channel to the end of the current montage.
        Take into account that the order of the channels is important!

        Parameters
        ----------
        channel: dict
            Dict with the channel data. The easiest way to calculate it is using
            function get_standard_channel_data_from_label. Keys:
                - label: channel label
                - coordinates: depends on the coordinate system and dimensions:
                        - Dim = 2D:
                            - Cartesian coordinates. Ex: {'x': 0.5, 'y': 0}
                            - Spherical coordinates Ex: {'r': 0.5,
                                'theta': np.pi/2}
                        - Dim = 3D:
                            - Cartesian coordinates. Ex: {'x': 0.5,
                                'y': 0, 'z': 0.8}
                            - Spherical coordinates. Ex: {'r': 0.5,
                                'theta': np.pi/2, 'phi': np.pi/4}
        reference: dict, list of dicts or str, optional
            For common and bipolar reference modes, reference must be a dict
            with the label and coordinates of the reference. For average
            reference, it can be 'all' to indicate a common average
            reference, or a list of dicts (with label and coordinates) of the
            averaged reference for each channel

        See Also
        --------
        get_standard_channel_data_from_label: returns channel data given the
            channel label and the standard. It can be used to get the reference
        """
        # TODO: check input
        channel['reference'] = reference
        channels = list() if self.channels is None else self.channels
        channels.append(channel)
        # Check channels
        self.__check_channels(channels)
        # Store attributes
        self.channels = channels
        self.n_cha = len(self.channels)
        self.l_cha = [cha['label'] for cha in self.channels]

    def set_montage(self, channels, ground=None,
                    allow_unlocated_channels=False):
        """Sets a custom montage, overwriting the previous one. Add single
        channels more easily using function add_custom_channel and
        add_standard_channel.

        Parameters
        ----------
        channels : list
            List of dicts, each of them representing a channel. The dict must
            contain the label, coordinates according to parameters dim an
            coord_system, and reference. For common reference mode, the
            reference must be a single channel with label and coordinates. For
            average reference mode, reference can be a list of dicts with the
            channels (label and coordinates) or "all", to specify common average
            reference. For bipolar reference mode, the reference must be a
            single channel with label and coordinates. In all cases you can set
            the reference to None, but this is not recommended. The definition
            of the coordinates may be skipped depending on the value of
            allow_unlocated_channels.
        ground : dict
            Dict containing the label and coordinates of the ground electrode
        allow_unlocated_channels: bool
            If False, the coordinates of all channels must be defined within
            channels dict. If True, the channels may not have known coordinates,
            This may be convenient if the localization of a channel is not known
            or it has a non-standard label, but the behaviour in functions that
            need coordinates (e.g., topographic_plots) is unpredictable.

        See Also
        --------
        set_standard_montage: preferred choice in most cases
        """
        # Check errors
        self.allow_unlocated_channels = allow_unlocated_channels
        self.__check_channels(channels, ground)
        # Set attributes
        self.channels = channels
        self.ground = ground
        self.n_cha = len(self.channels)
        self.l_cha = [cha['label'] for cha in self.channels]

    def set_standard_montage(self, l_cha=None, l_reference=None, l_ground=None,
                             montage='10-05', drop_landmarks=True,
                             allow_unlocated_channels=False, standard=None):
        """Set standard EEG channels with common reference. In 3 dimensions,
        the equator is taken a Nz-T10-Iz-T9.

        Parameters
        ----------
        l_cha : list, optional
            List of channels labels. The data will be returned keeping the
            same order. If None, the channels will be returned in the same order
            as they appear in the corresponding standard in medusa.meeg
        l_reference : str, optional
            Label of the reference. Usual choices are left preauricular point
            (LPA) and right preauricular point (RPA). Leave to None if you do
            not want to specify the reference (not recommended). Use only if
            reference_method is 'common'.
        l_ground : str, optional
            Label of the ground. Usual choices are AFz or FPz.
        montage : str {'10-20'|'10-10'|'10-05'} or dict
            EEG standard. If its a string, the corresponding labels and
            locations of the standard channels will be loaded using the files in
            medusa.meeg. To load a different montage, use function
            as meeg.read_montage_file and pass the returned dict here.
        drop_landmarks : bool
            Drop landmarks: nasion (NAS), left preauricular point (LPA) and
            right preauricular point (RPA) or mastoids (M1, M2). These are
            usually used as reference or ground points, so they are usually
            removed from the channel set for data analysis. Only used if l_cha
            is None.
        allow_unlocated_channels: bool
            If False, all the labels in parameter l_cha must be contained in the
            montage, which contains the corresponding coordinates. If True, the
            channels may not be in the standard, and they will be saved with no
            coordinates. This allows to save labels that are not defined in the
            montage, but the behaviour in functions that need locations
            (e.g., topographic_plots) is unpredictable.
        standard: str
            DEPRECATED. Only left for compatibility reasons.
        """
        # Check errors
        if self.reference_method != 'common':
            raise ValueError('Function set_standard_channels is available '
                             'only for recordings with common reference. For '
                             'custom montages use set_custom_channels')
        assert self.dim == '2D' or self.dim == '3D', \
            'Incorrect input on dim parameter'
        assert self.coord_system == 'cartesian' or \
               self.coord_system == 'spherical', \
            'Incorrect input on coord_system parameter'
        # Compatibility
        montage = standard if standard is not None else montage
        # Get montage
        if isinstance(montage, str):
            # Load standard montage
            self.montage = montage
            montage = meeg.get_standard_montage(standard=montage,
                                                dim=self.dim,
                                                coord_system=self.coord_system)
        else:
            # Set custom montage
            montage = montage.copy()
            self.montage = montage
        # Reference
        if l_reference is not None:
            if l_reference in montage:
                reference = montage[l_reference]
            else:
                if allow_unlocated_channels:
                    reference = dict()
                    warnings.warn('Reference not defined in montage')
                else:
                    raise meeg.ChannelNotFound(l_reference)
            reference['label'] = l_reference
        else:
            reference = None
        # Get list of labels to get
        labels = montage.keys() if l_cha is None \
            else [l.upper().strip() for l in l_cha]
        # Get channels
        channels = list()
        for label in labels:
            # Drop landmarks
            if l_cha is None:
                if drop_landmarks:
                    if label in ('NAS', 'LPA', 'RPA', 'M1', 'M2'):
                        continue
            # Append info
            if label in montage:
                channel_data = montage[label]
            else:
                if allow_unlocated_channels:
                    channel_data = dict()
                    warnings.warn('Channel %s not defined in montage' % label)
                else:
                    raise meeg.ChannelNotFound(label)
            channel_data['label'] = label
            channel_data['reference'] = reference
            channels.append(channel_data)
        # Ground
        if l_ground is not None:
            if l_ground in montage:
                ground = montage[l_ground]
            else:
                if allow_unlocated_channels:
                    ground = dict()
                    warnings.warn('Ground not defined in montage')
                else:
                    raise meeg.ChannelNotFound(l_ground)
            ground['label'] = ground
        else:
            ground = None
        # Check channels
        self.set_montage(channels, ground=ground,
                         allow_unlocated_channels=allow_unlocated_channels)

    def __check_channels(self, channels, ground=None):
        # Get mandatory and coordinates keys for each dim and coord_system
        cha_keys = ['label', 'reference']
        ref_keys = ['label']
        gnd_keys = ['label']
        if self.dim == '2D':
            if self.coord_system == 'cartesian':
                coord_keys = ['x', 'y']
            else:
                coord_keys = ['r', 'theta']
        else:
            if self.coord_system == 'cartesian':
                coord_keys = ['x', 'y', 'z']
            else:
                coord_keys = ['r', 'theta', 'phi']
        if not self.allow_unlocated_channels:
            cha_keys += coord_keys
            ref_keys += ref_keys
            gnd_keys += gnd_keys
        # Check keys
        references = list()
        for cha in channels:
            if not all(k in cha for k in cha_keys):
                raise ValueError('Malformed channel %s. Dict keys must be %s' %
                                 (str(cha), str(cha_keys)))
            # Check reference
            reference = cha['reference']
            references.append(reference)
            if reference is not None:
                if self.reference_method == 'common':
                    # Single reference
                    assert isinstance(reference, dict), \
                        'Reference must be None or dict'
                    if not all(k in reference for k in ref_keys):
                        raise ValueError('Malformed reference in channel %s. '
                                         'Reference keys must be %s' %
                                         (str(cha), str(ref_keys)))
                    # All references must be identical
                    if not all(ref == references[0] for ref in references):
                        raise ValueError('All references must be identical '
                                         'in common reference mode')
                elif self.reference_method == 'average':
                    # Average reference
                    assert isinstance(reference, list), \
                        'Reference must be None or list of dicts'
                    for r in reference:
                        if not all(k in r for k in ref_keys):
                            raise ValueError('Malformed reference in '
                                             'channel %s. Reference keys '
                                             'must be %s' %
                                             (str(cha), str(ref_keys)))
                elif self.reference_method == 'bipolar':
                    # Single reference
                    assert isinstance(reference, dict), \
                        'Reference must be of type dict'
                    if not all(k in reference for k in ref_keys):
                        raise ValueError('Malformed reference in channel %s. '
                                         'Reference keys must be %s' %
                                         (str(cha), str(ref_keys)))
        # Check ground
        if ground is not None:
            if not all(k in ground for k in gnd_keys):
                raise ValueError(
                    'Malformed ground. Dict keys must be %s' % (str(gnd_keys)))

    def get_cha_idx_from_labels(self, labels):
        """Returns the position of the channels given the labels

        Parameters
        ----------
         labels : list
            Labels to check. The order matters

        Returns
        -------
        indexes : np.ndarray
            Indexes of the channels in the set
        """
        return [self.l_cha.index(l) for l in labels]

    def check_channels_labels(self, labels, strict=False):
        """Checks the order and labels of the channels

        Parameters
        ----------
        labels : list
            Labels to check. The order matters
        strict : bool
            If True, comparison is strict. The function will check that the
            channel set contains the channels given by parameter labels and
            in the same order. If false, the function checks that the
            channels are contained in the channel set, but they could be in
            different order and the set could contain more channels

        Returns
        -------
        check : bool
            True if the labels and order are the same. False otherwise
        """
        if strict:
            check = True
            for i in range(len(labels)):
                if self.l_cha[i] != labels[i]:
                    check = False
        else:
            check = True
            for l in labels:
                if l not in self.l_cha:
                    check = False
        return check

    def subset(self, cha_idx):
        """Selects the channels given the indexes, creating a subset. The
        order of the channels will be updated

        Parameters
        ----------
        cha_idx : np.ndarray
            Indexes of the channels to select. The order matters
        """
        self.channels = [self.channels[idx] for idx in cha_idx]
        self.n_cha = len(self.channels)
        self.l_cha = [cha['label'] for cha in self.channels]

    def sort_nearest_channels(self):
        """ Sorts the nearest channels for each possible channel
        Return
        ----------
        sorted_dist_ch: dict()
            Dictionary that includes, for each channel (as a key), the rest
            of channels sorted by its closeness to that channel.
        """
        # TODO: Añadir la opción de hacer un sort de una coordenada específica,
        #  o en otra función
        if not self.channels:
            raise Exception(
                'Cannot compute the nearest channels if channel set '
                'is not initialized!')

        # Compute distance matrix
        dist_matrix = self.compute_dist_matrix()

        # Create the dictionary and format it
        sorted_dist_ch = dict()
        for i, label in enumerate(self.l_cha):
            d_sorted = np.sort(dist_matrix[i, :])
            idx_sorted = np.argsort(dist_matrix[i, :])
            if d_sorted[0] != 0:
                raise Exception('[meeg/EEGChannelSet/sort_nearest_channels] '
                                'Something ocurred, distance between the same '
                                'channel is not zero!')
            sorted_dist_ch[label] = []
            for j in range(1, len(d_sorted)):
                sorted_dist_ch[label].append({
                    "dist": d_sorted[j],
                    "channel": self.channels[idx_sorted[j]]
                })
        return sorted_dist_ch

    def compute_dist_matrix(self):
        """This function computes the distances between all channels in the
        channel set and stores them into a matrix.

        Returns
        -------------
        dist_matrix: ndarray of dimensions [ncha x ncha]
            Distances between all the channels.
        """
        if not self.channels:
            raise Exception(
                'Cannot compute the distance matrix if channel set '
                'is not initialized!')
        # Instantiate matrix of distances
        dist_matrix = np.empty((self.n_cha, self.n_cha))

        # For 2D coordinates
        if self.dim == '2D':
            if self.coord_system == 'cartesian':
                for i, cha in enumerate(self.channels):
                    # Find location of channel
                    cha_pos = np.array([cha['x'], cha['y']])
                    # Find the location of the rest of the channels
                    for j, temp_cha in enumerate(self.channels):
                        temp_cha_pos = np.array(
                            [temp_cha['x'], temp_cha['y']]
                        )
                        d = np.sqrt(np.sum(np.power(temp_cha_pos - cha_pos, 2)))
                        dist_matrix[i, j] = d
            elif self.coord_system == 'spherical':
                for i, cha in enumerate(self.channels):
                    # Find location of channel
                    r_cha, theta_cha = cha['r'], cha['theta']
                    # Find the location of the rest of the channels
                    for j, temp_cha in enumerate(self.channels):
                        r_temp_cha, theta_temp_cha = \
                            temp_cha['r'], temp_cha['theta']
                        d = np.abs(np.sqrt(r_cha ** 2 + r_temp_cha ** 2 -
                                           2 * r_cha * r_temp_cha *
                                           np.cos(theta_temp_cha -
                                                  theta_cha)))
                        dist_matrix[i, j] = d
        # For 3D coordinates
        elif self.dim == '3D':
            if self.coord_system == 'cartesian':
                for i, cha in enumerate(self.channels):
                    # Find location of channel
                    cha_pos = np.array([cha['x'], cha['y'], cha['z']])
                    # Find the location of the rest of the channels
                    for j, temp_cha in enumerate(self.channels):
                        temp_cha_pos = np.array(
                            [temp_cha['x'], temp_cha['y'], temp_cha['z']]
                        )
                        d = np.sqrt(np.sum(np.power(temp_cha_pos - cha_pos, 2)))
                        dist_matrix[i, j] = d
            elif self.coord_system == 'spherical':
                for i, cha in enumerate(self.channels):
                    # Find location of channel
                    r_cha, theta_cha, phi_cha = \
                        cha['r'], cha['theta'], cha['phi']
                    # Find the location of the rest of the channels
                    for j, temp_cha in enumerate(self.channels):
                        r_temp_cha, theta_temp_cha, phi_temp_cha = \
                            temp_cha['r'], temp_cha['theta'], temp_cha['phi']
                        d = np.abs(
                            np.sqrt(
                                r_cha ** 2 +
                                r_temp_cha ** 2 -
                                2 * r_cha * r_temp_cha *
                                (
                                        np.sin(theta_cha) *
                                        np.sin(theta_temp_cha) *
                                        np.cos(phi_cha - phi_temp_cha) +
                                        np.cos(theta_temp_cha) *
                                        np.cos(theta_cha))
                            )
                        )
                        dist_matrix[i, j] = d
        return dist_matrix

    def to_serializable_obj(self):
        return self.__dict__

    @classmethod
    def from_serializable_obj(cls, dict_data):
        inst = cls()
        inst.__dict__.update(dict_data)
        return inst


class MEGChannelSet(components.SerializableComponent):

    # TODO
    def __init__(self):
        self.channels = None
        self.n_cha = None
        self.l_cha = None

    def to_serializable_obj(self):
        return self.__dict__

    @classmethod
    def from_serializable_obj(cls, dict_data):
        inst = cls()
        inst.__dict__.update(dict_data)
        return inst


class EEG(components.BiosignalData):
    """Electroencephalography (EEG) biosignal
    """

    def __init__(self, times, signal, fs, channel_set, **kwargs):
        """EEG constructor

        Parameters
        ----------
        times : list or numpy.ndarray
            1D numpy array [n_samples]. Timestamps of each sample. If they are
            not available, generate them
            artificially. Nevertheless, all signals and events must have the
            same temporal origin
        signal : list or numpy.ndarray
            2D numpy array [n_samples x n_channels]. EEG samples (the units
            should be defined using kwargs)
        fs : int or float
            Sample rate of the recording.
        channel_set : meeg_standards.EEGChannelSet
            EEG channel set
        kwargs: kwargs
            Any other parameter provided will be saved in the class (e.g.,
            equipment description)
        """
        # To numpy arrays
        times = np.array(times)
        signal = np.array(signal)
        # Check errors
        if signal.shape[1] != channel_set.n_cha:
            raise Exception("Signal with shape [samples x channels] does not "
                            "match with the number of channels")
        if times.shape[0] != signal.shape[0]:
            raise Exception("Parameters times (shape: %s) and signal (shape: "
                            "%s) must have the same length" % (
                                str(times.shape), str(signal.shape))
                            )

        # Standard attributes
        self.times = times
        self.signal = signal
        self.fs = fs
        self.channel_set = channel_set

        # Optional attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def change_channel_set(self, channel_set):
        """Smart change of channel set, updating the signal and all related
        attributes

        Parameters
        ----------
        channel_set : meeg_standards.EEGChannelSet
            EEG channel set
        """
        # Get the index of the channels
        cha_idx = self.channel_set.get_cha_idx_from_labels(channel_set.l_cha)
        # Select and reorganize channels channels
        self.channel_set.subset(cha_idx)
        # Reorganize signal
        self.signal = self.signal[:, cha_idx]

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
            if type(rec_dict[key]) == EEGChannelSet:
                rec_dict[key] = rec_dict[key].to_serializable_obj()
        return rec_dict

    @classmethod
    def from_serializable_obj(cls, dict_data):
        # Load channel set dict
        dict_data['channel_set'] = EEGChannelSet.from_serializable_obj(
            dict_data['channel_set']
        )
        return cls(**dict_data)


class MEG(components.BiosignalData):
    # TODO check everything

    """Magnetoencephalography (MEG) biosignal
    """

    def __init__(self, times, signal, fs, channel_set, **kwargs):
        """MEG constructor

        Parameters
        ----------
        times : list or numpy.ndarray
            1D numpy array [n_samples]. Timestamps of each sample. If they are
            not available, generate them
            artificially. Nevertheless, all signals and events must have the
            same temporal origin
        signal : list or numpy.ndarray
            2D numpy array [n_samples x n_channels]. MEG samples (the units
            should be defined using kwargs)
        fs : int or float
            Sample rate of the recording.
        channel_set : list or meeg_standards.MEGChannelSet
            MEG channel set.
        kwargs: kwargs
            Any other parameter provided will be saved in the class (e.g.,
            equipment description)
        """
        # To numpy arrays
        times = np.array(times)
        signal = np.array(signal)
        # Check errors
        if signal.shape[1] != channel_set.n_cha:
            raise Exception("Signal with shape [samples x channels] does not "
                            "match with the number of channels")
        if times.shape[0] != signal.shape[0]:
            raise Exception("Parameters times and signal must have the same "
                            "length")

        # Standard attributes
        self.times = times
        self.signal = signal
        self.fs = fs
        self.channel_set = channel_set

        # Optional attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def change_channel_set(self, channel_set):
        """Smart change of channel set, updating the signal and all related
        attributes

        Parameters
        ----------
        channel_set : meeg_standards.MEGChannelSet
            MEG channel set
        """
        raise NotImplementedError

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
        return rec_dict

    @classmethod
    def from_serializable_obj(cls, dict_data):
        return cls(**dict_data)

    @staticmethod
    def load_meg_signal_from_spm_file(path, fs):
        # TODO check everything

        """Function to load a MEG recording from a spm file

        Parameters
        ----------
        path : str
            Path of the file
        fs : int or float
            Sample rate of the recording
        """
        data = scipy.io.loadmat(path)
        info = data['D'][0, 0]

        datatype = info['datatype'][0]
        num_chan = np.size(info['channels'])
        num_samples = info['Nsamples'][0, 0]

        raw = np.fromfile(path[0:len(path) - 3] + 'dat',
                          datatype[0:len(datatype) - 3])
        signal = np.reshape(raw, [num_chan, num_samples], order='F')
        times = np.linspace(0, signal[0] / fs, signal[0])
        channels = None
        return MEG(times, signal, fs, channels)


class Connecitivity:
    # TODO: check everything

    """Customizable class with connectivity info from EEG/MEG recordings
    """

    def __init__(self, data, trial_len, parameter, filt_mode, **kwargs):
        """Class constructor

        Parameters
        ----------
        data : bla bla
            Bla bla
        trial_len : bla bla
            Bla bla
        parameter : bla bla
            Bla bla
        filt_mode : bla bla
            Bla bla
        kwargs
            Optional information of the EEG recording (e.g. subject, amplifier,
            etc)
        """
        # Params
        data = np.array(data)

        if not (filt_mode == 'all' or
                filt_mode == 'bands' or
                filt_mode == 'win'):
            raise ValueError("Unknown filtering mode")

        self.data = data
        self.trial_len = trial_len
        self.parameter = parameter
        self.filt_mode = filt_mode

        # Optional attributes
        for key, value in kwargs.items():
            setattr(self, key, value)


class UnknownStandardChannel(Exception):

    def __init__(self, msg=None):
        """Class constructor

        Parameters
        ----------
        msg: string or None
            Custom message
        """
        if msg is None:
            msg = 'Unknown standard channel'
        super().__init__(msg)


import tensorflow as tf
import numpy as np


def __aux_symm_triu_gpu(W):
    N = tf.shape(W)
    aux = tf.subtract(tf.linalg.band_part(tf.ones(N), 0, -1),tf.linalg.band_part(tf.ones(N), 0, 0))
    W = tf.math.multiply(W,tf.cast(aux,tf.float64))
    W = tf.math.add(W,tf.transpose(W))
    W = tf.math.divide(tf.math.reduce_sum(W,axis=0),2)
    return W


def __aux_no_match_gpu(W):
    in_degree = tf.math.reduce_sum(W,axis=0)
    out_degree = tf.math.reduce_sum(W,axis=1)
    W = tf.math.add(in_degree, out_degree)
    return W    


def __degree_gpu(W):
    """
    Calculates node degree (also called strength in weighted networks) using GPU

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    nodal_degree : numpy array
        Nodal degree.      

    """
    W = tf.math.divide(tf.math.round(tf.math.multiply(W,10000000000)),10000000000)
    W = W - tf.cast(tf.eye(tf.shape(W)[0]),tf.float64)    
    
    check_symmetry = tf.reduce_all(tf.math.equal(W,tf.transpose(W)))
    check_symmetry = tf.cond(
        tf.math.reduce_sum(tf.subtract(tf.linalg.band_part(W, -1, 0),tf.linalg.band_part(W, 0, 0))) == 0,
        lambda: 1, lambda: check_symmetry)
    check_symmetry = tf.cond(
        tf.reduce_all(tf.math.equal(W,-tf.transpose(W))),
        lambda: 2, lambda: check_symmetry)
        
    nodal_degree = tf.switch_case(tf.cast(check_symmetry,tf.int32), 
                branch_fns={0: lambda: __aux_no_match_gpu(W), 1: lambda: __aux_symm_triu_gpu(W), 2: lambda: -tf.math.reduce_sum(W,axis=0)})     
      
    return nodal_degree


def __aux_symm_triu_cpu(W):
    N = np.shape(W)[0]
    aux = np.ones((N,N))
    aux = np.triu(aux,k=1)
    W = W * aux
    W = W + np.transpose(W)
    W = np.sum(W,axis=0) / 2
    return W


def __aux_no_match_cpu(W):
    in_degree = np.sum(W,axis=0)
    out_degree = np.sum(W,axis=1)
    W = in_degree + out_degree
    return W    


def __degree_cpu(W):
    """
    Calculates node degree (also called strength in weighted networks) using CPU

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    nodal_degree : numpy array
        Nodal degree.      

    """
    W = np.divide(np.round(W * 10000000000),10000000000)
    W = W - np.diag(np.diag(W))    
    
    check_symmetry = (W.transpose() == W).all() # if symmetric
    
    if (W == np.triu(W)).all(): # if upper triangular
        check_symmetry = 1
        
    if (W.transpose() == -W).all(): # if anti-symmetric
        check_symmetry = 2
        
    if check_symmetry == 0:
        nodal_degree = __aux_no_match_cpu(W)
    elif check_symmetry == 1:
        nodal_degree =  __aux_symm_triu_cpu(W)
    elif check_symmetry == 2:
        nodal_degree = -np.sum(W,axis=0)
          
    return nodal_degree


def degree(W,mode):
    """
    Calculates the graph degree.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    nodal_degree : numpy array
        Nodal degree.      

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    if mode == 'CPU':
        nodal_degree = __degree_cpu(W)
    elif mode == 'GPU':
        nodal_degree = __degree_gpu(W)
    else:
        raise ValueError('Unknown mode')

    return nodal_degree


import numpy as np

def modularity(W,gamma):
    """
    Calculates the optimal node comunity and the modularity
    
    Note: The optimal community structure subdivides the network in groups of 
    nodes (non-overlapping), maximizing withing-groups edges while minimising 
    between-groups edges. Modularity quantifies the degree to which the graph 
    could be subdivided into the aforementioned communities
        
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    gamma : float
        modularity resolution parameter     
            gamma>1     smaller modules
            0<=gamma<1  larger modules
            gamma=1     (default) classic modularity function
        
    Returns
    -------
    Q : int
        Network modularity. 
    Ci : numpy array
        Network communities.
        
    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')  
    
    if 'gamma' not in locals():
        gamma = 1 # Default
        
    N = W.shape[0]
    n_perm = np.random.permutation(N) - 1
    W = W[:, n_perm][n_perm] # Randomly permute the order of the nodes
    K = np.sum(W,axis=0) # Degree
    K = K[np.newaxis,:]
    m = np.sum(K) # # of edges (undirected edges are counted twice)
    B = W - gamma * (np.matmul(K.T,K)) / m # Modularity matrix
    Ci = np.ones((N,1)) # Community indices
    cn = 1 # Number of communities
    U = np.array([1, 0]) # Array of unexamined communities
    
    ind = np.arange(N)
    ind = ind[:,np.newaxis]
    Bg = np.copy(B)
    Ng = N
    
    while U[0]: # Community U[0]
        D,V = np.linalg.eig(Bg)
        il = np.argmax(np.real(D))
        v1 = V[:,il] # Eigenvector of the max eigenvalue of Bg
        
        S = np.ones((Ng,1))
        S[v1<0,0] = -1
        q = np.matmul(np.matmul(S.T,Bg),S) # Contribution to modularity

        if q > 10**-10: # Contribution positive: U[0] is divisible
            qmax = q # Max contribution to modularity
            Bg[np.eye(Ng,dtype=np.bool)] = 0 # To enable fine-tuning, modify Bg
            indg = np.ones((Ng,1)) # Array of unmoved indices
            Sit = np.copy(S)
            while not (np.isnan(indg)).all(): # Iterative fine-tuning
                Qit = qmax - 4 * Sit * np.matmul(Bg,Sit)
                qmax = np.nanmax(Qit * indg)
                imax = (Qit == qmax)
                Sit[imax[:,0],0] = -Sit[imax[:,0],0]
                indg[imax[:,0],0] = np.nan
                if qmax > q:
                    q = qmax
                    S = np.copy(Sit)
                    
            if abs(np.sum(S)) == Ng: # Unsuccessful splitting of U[0]
                U[0] = []
            else:
                cn = cn + 1
                Ci[ind[S[:,0]==1][:,0],0] = U[0] # Split old U[0] into new U[0] and into cn
                Ci[ind[S[:,0]==-1][:,0],0] = cn
                U = np.insert(U,0,cn)
        else: # Contribution nonpositive: U[0] is indivisible
            U = np.delete(U, 0)
            
        ind = np.argwhere(Ci == U[0]) # Indices of unexamined community U[0]
        bg = B[:,ind[:,0]][ind[:,0]]
        Bg = bg - np.diag(np.sum(bg,axis=0)) # Modularity matrix for U[0]
        Ng = ind.shape[0] # Number of vertices in U[0]

    s = np.tile(Ci, (1, N)) # Compute modularity
    Q = (s - s.T)
    Q[Q!=0] = 1
    Q = np.logical_not(Q).astype(int) * B / m
    Q = np.sum(np.squeeze(np.asarray(Q)))
    Ci_corrected = np.zeros((N,1)) # Initialise Ci_corrected
    Ci_corrected[n_perm,0] = Ci[:,0] # Return order of nodes to that used in the input stage
    Ci = Ci_corrected # Output corrected community assignment

    return Q,Ci



import numpy as np

def betweenness(W,norm = True):
    """
    Calculates the betweenness centrality of the graph, which is the probability 
    that node i belong to one of the network shortest paths
    
    Note: W must be converted to a connection-length matrix. It is common to 
    obtain it via mapping from weight to length. BC is normalised to the 
    range [0 1] as BC/[(N-1)(N-2)]
        
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    norm : bool
        Normalisation. 0 = no, 1 = yes. Default = True

    Returns
    -------
    BC : numpy array
        Network betweenness centrality.
        
    """
    if not (norm == 0 or norm == 1 or norm == True or norm == False):
        raise ValueError('Variable "norm" must be either 0, 1, or bool')
    # norm = norm.astype(bool)        
        
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')      
    
    n = W.shape[0]
    E = np.argwhere(W)
    W[E] = 1/W[E] # Invert weights
    BC = np.zeros((n,1)) # Vertex betweenness
    
    for u in range(0,n):
        D = np.ones((1,n)) * np.inf
        D[0,u] = 0 # Distance from u
        NP = np.zeros((1,n))
        NP[0,u] = 1 # Number of paths from u
        S = np.ones((1,n)) # Distance permanence (1 is not permanent)
        P = np.zeros((n,n)) # Predecessors
        Q = np.zeros((1,n))
        q = n - 1# Order of non-increasing distance
        
        W1 = np.copy(W)
        V = [u]
        
        while 1:
            S[0,V] = 0 # Distance u>V is now permanent
            W1[:,V] = 0 # No in-edges as already shortest
            for v in V:
                Q[0,q] = v + 1
                q = q - 1
                G = np.argwhere(W1[v,:]) # Neighbours of v
                for g in G:
                    Duw = D[:,v] + W1[v,g] # Path length to be tested
                    
                    if Duw < D[0,g]: # If new u>W shorter than old
                        D[0,g] = Duw
                        NP[0,g] = NP[0,v] # NP(u>w) = NP of new path
                        P[g,:] = 0
                        P[g,v] = 1 # v is the only predecessor
                    elif Duw == D[0,g]: # if new u>v equal to old
                        NP[0,g] = NP[0,g] + NP[0,v] # NP(u>w) sum of old and new
                        P[g,v] = 1 # v is also a predecessor
             
            if not D[:,np.squeeze(S).astype(bool)].any(): # All nodes reached...
                break
            
            minD = np.min(D[:,np.squeeze(S).astype(bool)])

            if minD == np.inf: # ...some cannot be reached:
                Q[:,0:q+1] = np.transpose(np.argwhere(np.squeeze(D) == np.inf)) # ...these are first-in-line
                break
            
            V = np.argwhere(D==minD)
            V = V[:,1]
            
        DP = np.zeros((n,1)) # Dependency
        for w in (Q[0,0:n-1]-1).astype(int):
            BC[w,0] = BC[w,0] + DP[w,0]
            for v in np.argwhere(P[w,:]):
                DP[v,0] = DP[v,0] + (1 + DP[w,0]) * NP[0,v] / NP[0,w]
    if norm:
        BC = BC / ((n - 1) * (n - 2))
                
    return BC            
        


import numpy as np
from scipy import signal


def get_epochs(signal, epochs_length, stride=None, norm=None):
    """This function returns the signal divided in epochs following a sliding
    window approach

   Parameters
   ----------
   signal : list or numpy.ndarray
       Array to extract epochs with shape [samples, channels]
   epochs_length : int
       Epochs length in samples
   stride : int, optional
       Separation between consecutive epochs in samples. If None, stride is set
       to epochs_length.
   norm : str, optional
       Set to 'z' for Z-score normalization or 'dc' for DC normalization.
       Statistical parameters are computed using the whole epoch.

   Returns
   -------
   numpy.ndarray
       Structured data with dimensions [n_epochs x length x n_channels]
    """
    # Check errors
    if epochs_length <= 0:
        ValueError('Parameter epochs_length must be greater than 0')
    if stride is None:
        stride = epochs_length
    if stride <= 0:
        raise ValueError('Parameter stride must be None or greater than 0')
    # Assure ints
    n_cha = signal.shape[1]
    epochs_length = int(epochs_length)
    stride = int(stride)
    # Get epochs
    epochs = np.lib.stride_tricks.sliding_window_view(
        signal, (epochs_length, n_cha))[::stride].squeeze()
    # Normalize
    if norm is not None:
        epochs = normalize_epochs(epochs, norm)
    return epochs


def get_epochs_of_events(timestamps, signal, onsets, fs, w_epoch_t,
                         w_baseline_t=None, norm=None):

    """This function calculates the epochs of signal given the onsets of events.

    Parameters
    ----------
    timestamps : list or numpy.ndarray
        Timestamps of each biosignal sample
    signal : list or numpy.ndarray
        Biosignal samples
    onsets : list or numpy.ndarray
        Events timestamps
    fs : float
        Sample rate
    w_epoch_t : list or numpy.ndarray
        Temporal window in ms of the epoch. For example, t_baseline = [0, 1000]
        takes the epoch form 0 ms to 1000 ms after each onset (0 ms represents
        the onset).
    w_baseline_t : list or numpy.ndarray, optional
        Temporal window in ms of the baseline. For example, t_baseline =
        [-500, 100] takes the baseline form -500 ms before each onset to 100
        ms after each onset (0 ms represents the onset). This chunk of signal is
        used to normalize the epoch, if aplicable
    norm : str, optional
       Set to 'z' for Z-score normalization or 'dc' for DC normalization.
       Statistical parameters are computed using the whole epoch.
    Returns
    -------
    numpy.ndarray
        Structured data with dimensions [events x samples x channels]
    """
    # Error prevention
    epoch_feasibility = check_epochs_feasibility(timestamps, onsets, fs,
                                                 w_epoch_t)
    if epoch_feasibility == 'first':
        raise ValueError("Not enough EEG samples to get the first epoch")
    elif epoch_feasibility == 'last':
        raise ValueError("Not enough EEG samples to get the last epoch")
    if w_baseline_t is not None:
        baseline_feasibility = check_epochs_feasibility(timestamps, onsets, fs,
                                                        w_baseline_t)
        if baseline_feasibility == 'first':
            raise ValueError("Not enough EEG samples to get the first baseline")
        elif baseline_feasibility == 'last':
            raise ValueError("Not enough EEG samples to get the last baseline")
        if norm is None:
            raise ValueError('If parameter w_baseline_t is not None, please '
                             'specify the normalization type with parameter '
                             'norm')
    if norm is not None and w_baseline_t is None:
        raise ValueError('If parameter norm is not None, please specify the '
                         'baseline window with parameter w_baseline_t')

    # Useful parameters
    w_epoch_t = np.array(w_epoch_t)
    w_epoch_s = np.array(w_epoch_t * fs / 1000, dtype=int)
    l_epoch = w_epoch_s[1] - w_epoch_s[0]
    # For each onset
    n_cha = signal.shape[1]
    epochs_idx = np.searchsorted(timestamps, onsets + w_epoch_t[0] / 1000,
                                 side='left')
    epochs = np.lib.stride_tricks.sliding_window_view(
        signal, (l_epoch, n_cha))[epochs_idx].squeeze(axis=1)
    # Baseline normalization
    if w_baseline_t is not None and norm is not None:
        # Baseline start-end (samples)
        w_baseline_t = np.array(w_baseline_t)
        w_baseline_s = np.array(w_baseline_t * fs / 1000, dtype=int)
        l_baseline = w_baseline_s[1] - w_baseline_s[0]
        # Extract baselines
        baseline_idx = np.searchsorted(
            timestamps, onsets + w_baseline_t[0] / 1000, side='left')
        baselines = np.lib.stride_tricks.sliding_window_view(
            signal, (l_baseline, n_cha))[baseline_idx].squeeze(axis=1)
        epochs = normalize_epochs(epochs, norm_epochs=baselines, norm=norm)
    return epochs


def normalize_epochs(epochs, norm_epochs=None, norm='z'):
    """
    Normalizes epochs

    Parameters
    ----------
    epochs: list or numpy.ndarray
        Epochs of signal with dimensions [n_epochs x n_samples x n_channels]
    norm_epochs: list or numpy.ndarray, optional
        Epochs of signal with dimensions [n_epochs x n_samples x n_channels]
        that are used to compute the statistical parameters for normalization.
        If None, norm_epochs=epochs.
    norm: str
        Set to 'z' for Z-score normalization or 'dc' for DC normalization.
        Statistical parameters are computed using the whole epoch.
    """
    # Check errors
    if norm_epochs is None:
        norm_epochs = epochs
    if norm not in ['z', 'dc']:
        raise ValueError("Parameter norm must be 'z' or 'dc'")
    # Normalization
    if norm == 'z':
        # z-score
        mean = np.mean(norm_epochs, axis=1, keepdims=True)
        std = np.std(norm_epochs, axis=1, keepdims=True)
        epochs = (epochs - mean) / std
    elif norm == 'dc':
        # DC subtraction
        mean = np.mean(norm_epochs, axis=1, keepdims=True)
        epochs = epochs - mean
    else:
        raise ValueError("Parameter norm must be 'z' or 'dc'")
    return epochs


def resample_epochs(epochs, t_window, target_fs):
    """Resample epochs to the target_fs.

    IMPORTANT: No antialising filter is applied

    Parameters
    ----------
    epochs : list or numpy.ndarray
        Epochs of signal with dimensions [n_epochs x samples x channels]
    t_window : list or numpy.ndarray
        Temporal window in ms of the epoch. For example, t_window = [0, 1000]
        takes the epoch form 0 ms to 1000 ms after each onset (0 ms
        represents the onset).
    target_fs : float
        Target sample rate

    Returns
    -------
    numpy.ndarray
        Final epochs with dimensions [events x target_samples x channels]
    """
    # Compute the desired window and baseline in samples
    l_t_window = t_window[1] - t_window[0]  # Window length (ms)
    target_samples = np.floor((target_fs * l_t_window) / 1000).astype(int)
    # Resample epochs
    epochs = signal.resample(epochs, target_samples, axis=1)
    return epochs


def check_epochs_feasibility(timestamps, onsets, fs, t_window):
    """Checks if the extraction of the desired window is feasible with the
    available samples. Sometimes, the first/last stimulus sample is so close
    to the beginning/end of the signal data chunk that there are not enough
    samples to compute this window. In this case, the function will return
    "first" or "last" to identify which onset can not be extracted with
    the current t_window. It will return 'ok' if there is no conflict.

    Parameters
    ----------
     timestamps : list or numpy.ndarray
        Timestamps of each biosginal sample
    onsets : list or numpy.ndarray
        Events timestamps
    fs : float
        Sample rate
    t_window : list or numpy.ndarray
        Temporal window in ms of the epoch. For example, t_window = [0, 1000]
        takes the epoch form 0 ms to 1000 ms after each onset (0 ms represents
        the onset).

    Returns
    -------
    feasibility : string
        "ok" If window extraction is feasible.
        "first" If window extraction is not feasible for the first onset.
        "last" If window extraction is not feasible for the last onset.
    """
    first_sti_sam_onset = np.argmin(np.abs(timestamps - onsets[0]))
    last_sti_sam_onset = np.argmin(np.abs(timestamps - onsets[-1]))
    s_window = np.array(np.array(t_window) * fs / 1000, dtype=int)
    last_sti_sam_end = last_sti_sam_onset + s_window[1]
    first_sti_sam_beginning = first_sti_sam_onset + s_window[0]

    if first_sti_sam_beginning < 0:
        feasibility = 'first'
    elif last_sti_sam_end >= len(timestamps):
        feasibility = 'last'
    else:
        feasibility = 'ok'
    return feasibility


def reject_noisy_epochs(epochs, signal_mean, signal_std, k=4, n_samp=2,
                        n_cha=1):
    """Simple thresholding method to reject noisy epochs. It discards epochs
    with n_samp samples greater than k*std in n_cha channels

    Parameters
    ----------
     epochs : list or numpy.ndarray
        Epochs of signal with dimensions [n_epochs x samples x channels]
    signal_mean : float
        Mean of the signal
    signal_std : float
        Standard deviation of the signal
    k : float
        Standard deviation multiplier to calculate threshold
    n_samp : int
        Minimum number of samples that have to be over the threshold in each
        epoch to be discarded
    n_cha : int
        Minimum number of channels that have to have n_samples over the
        threshold in each epoch to be discarded

    Returns
    -------
     float
        Percentage of reject epochs in
    numpy.ndarray
        Clean epochs
    numpy.ndarray
        Indexes for rejected epochs. True for discarded epoch
    """

    # Check errors
    if len(epochs.shape) != 3:
        raise Exception('Malformed epochs array. It must be of dimmensions '
                        '[epochs x samples x channels]')
    if signal_std.shape[0] != epochs.shape[2]:
        raise Exception('Array signal_std does not match with epochs size. '
                        'It must have the same number of channels')
    if signal_mean.shape[0] != epochs.shape[2]:
        raise Exception('Array signal_mean does not match with epochs size. '
                        'It must have the same number of channels')

    epochs_abs = np.abs(epochs)
    cmp = epochs_abs > np.abs(signal_mean) + k * signal_std
    idx = np.sum((np.sum(cmp, axis=1) >= n_samp), axis=1) >= n_cha
    pct_rejected = (np.sum(idx) / epochs.shape[0]) * 100
    return pct_rejected, epochs[~idx, :, :], idx


def time_to_sample_index_events(times, onsets):
    """Converts an array of time onsets to an array that indicates the sample
    index of the event

        Parameters
        ----------
        times : list or numpy.ndarray
            Array of shape [n_samples]. Timestamps of the biosignal
        onsets : list or numpy.ndarray
            Array of shape [n_events]. Onsets in time of the events

        Returns
        -------
        numpy.ndarray
            Array of shape [n_samples]. The array is 1 only in the nearest
            timestamp to each onset, otherwise 0
        """

    # Check errors
    times = np.squeeze(times)
    onsets = np.squeeze(onsets)

    if len(times.shape) != 1:
        raise ValueError('Parameter times must be a 1D array')
    if len(onsets.shape) != 1:
        raise ValueError('Parameter onsets must be a 1D array')

    rep_times = np.matlib.repmat(times, onsets.shape[0], 1).T
    rep_onsets = np.matlib.repmat(onsets, times.shape[0], 1)

    return np.argmin(np.abs(rep_times - rep_onsets), axis=0)
